# -*- encoding: utf-8 -*-
"""
hio.core.doing Module
"""
import time
import types
import inspect
from inspect import isgeneratorfunction
from collections import deque, namedtuple

from .. import hioing
from .basing import State
from . import tyming
from ..help import timing, helping


Deed = namedtuple("Deed", "dog retyme doer")

class Doist(tyming.Tymist):
    """
    Doist is the root coroutine scheduler
    Provides relative cycle time in seconds with .tyme property to doers it runs
    The relative cycle time is advanced in .tock size increments by the  by  the
    .tick method.
    The doist may treat .tyme as artificial time or synchonize it to real time.

    .enter method prepares deeds deque of triples (dog, retyme, doer) where
        dog is a doer generator returned by calling doer generator instances,
        functions, or methods.

    .recur method runs its deeds deque of triples (dog, retyme, doer) once per
        invocation.
        This synchronizes their cycle time .tyme to the Doist's tyme.

    .do method repeatedly runs .recur until generators are complete
       it may either repeat as fast as possbile or repeat at real time increments.

    Inherited Class Attributes:
        .Tock is default .tock

    Inherited Attributes:


    Attributes:
        name (str): unique identifier of doist uses for identifying resources of
                    doists running in child processes in multiprocessing
        real (bool): True means run in real time, Otherwise as fast as possible.
        limit (float):  maximum run tyme limit then closes all doers
        done (bool | None): True means completed due to limit or all deeds completed
                False is forced complete due to error
        doers (list): Doer class instances, generator methods or
                function callables with attributes tock, done, and opts dict().
                Used throughout the execution lifecycle.
        deeds (deque): Tuples of form (dog, retyme, doer). Where:
            dog is generator created by doer
            retyme is tyme (real or simulated) in seconds when dog should run next
            doer is associated doer in .doers list used to assign its .done state
                given completion state of its dog
            Used throughout the execution lifecycle. The normal
            case is use the default empty initialization performed here and
            update in .enter().
        timer (MonoTimer): for real time intervals
        temp (bool): True means use temp resources such as file path.
                     When True inject into doer enters when True.
                     Otherwise do not inject into doer enters.

    Inherited Properties:
        tyme: (float): starting relative cycle time, .tyme is artificial time
        tock (float | None): float tyme lag of .tick(). None means asap

    Properties:

    Inherited Methods:
        .tick increments .tyme by one .tock or provided tock

    Methods:
        .enter prepare deeds, deque of triples (dog, retyme, doer)
        .recur  run through all deeds once
        .do repeadedly call .recur until all dogs in deeds are complete or
            times out do to reaching time limit

    """
    def __init__(self, *, name='doist', real=False, limit=None, doers=None,
                          temp=False, **kwa):
        """
        Returns:
            instance

        Inherited Parameters:
            tyme (float): initial value of cycle time in seconds
            tock (float | None): lag tyme in seconds between runs, None means run ASAP

        Parameters:
            name (str): unique identifier of doist to manage resources
            real (boolean): True means run in real time,
                            Otherwise run faster than real
            limit (float): seconds for max run time of doist. None means no limit.
            doers (iterable): Doer class instances, generator methods or
                function callables with attributes tock, done, and opts dict()
                used to initialize .doers.
                The .doers attribute is used throughout the execution lifecycle.
                Parameterization elsewhere of doers enables some special cases.
                The normal case is to initialize here or in .do().
            temp (bool): True means use temp resources such as file path, inject
                         into doers when True. Otherwise do not inject.
        """
        super(Doist, self).__init__(**kwa)
        self.name = name
        self.real = True if real else False
        self.limit = abs(float(limit)) if limit is not None else None
        self.done = None
        self.doers = list(doers) if doers is not None else []  # list of Doers
        self.deeds = deque()  # deque of deeds
        self.timer = timing.MonoTimer(duration = self.tock)
        self.temp = True if temp else False


    def do(self, doers=None, limit=None, tyme=None, *, temp=None):
        """
        Readies deeds deque from .doers or doers if any and then iteratively
        runs .recur over deeds deque until completion of all deeds.
        Each entry in deeds is a triple (dog, retyme, doer)  where:
            dog is generator
            retyme is tyme (real or simulated) in seconds when dog should run next
            doer is from .doers list used to assign its .done state given
            associated completion state of its dog

        If interrupted by exception call .close on each dog to force exit context.

        Keyboard interrupt (cntl-c) forces exit.

        Once finally clause closes a generator it must be reinited
        before it can be run again

        Parameters:
            doers (iterable): generator method or function callables with attributes
                tock, done, and opts dict(). This may be used to update the .doers
                attribute which is used throughout the execution lifecycle.
                If not provided uses .doers.
                Parameterization here of doers enables some special cases.
                The normal case is to initialize in .__init__ or here.
            limit (float): is real time limit on execution. Forces close of all dogs.
            tyme  (float): is optional starting tyme. Resets .tyme to tyme whe provided.
               If not provided uses current .tyme
               temp (bool): True means use temp resources such as file path, inject
                         into doers when True. Otherwise do not inject.

        Returns:
            None

        See: https://stackoverflow.com/questions/40528867/setting-attributes-on-func
        For setting attributes on bound methods.
        """
        temp = temp or (self.temp if self.temp else temp)  # inject if temp or self.temp

        self.done = False
        if doers is not None:
            self.doers = list(doers)
            self.deeds = deque()

        if limit is not None:  # time limt for running if any. useful in test
            self.limit = abs(float(limit))

        if tyme is not None:  # re-initialize starting tyme
            self.tyme = tyme

        try:  # always clean up resources upon exception
            self.enter(temp=temp)  # runs enter context on each doer

            tymer = tyming.Tymer(tymth=self.tymen(), duration=self.limit)
            self.timer.start()

            while True:  # until doers complete or exception or keyboardInterrupt
                try:
                    self.recur()  # increments .tyme runs recur context

                    if self.real:  # wait for real time to expire
                        while not self.timer.expired:
                            time.sleep(max(0.0, self.timer.remaining))
                        self.timer.restart()  #  no time lost

                    if not self.deeds:  # no deeds
                        self.done = True
                        break  # break out of forever loop

                    if self.limit and tymer.expired:  # reached limit before all deeds done
                        break  # break out of forever loop

                except KeyboardInterrupt:  # Forced shutdown due to SIGINT, use CNTL-C to shutdown from shell
                    break

                except SystemExit: # Forced shutdown of process via sys.exit()
                    raise

                except Exception:  # Forced shutdown due to uncaught exception
                    raise

        finally: # finally clause always runs regardless of exception or not.
            self.exit()  # force close remaining deeds throws GeneratorExit


    def enter(self, doers=None, *, temp=None):
        """Enter context
        Returns (deque):  deeds deque of triples (dog, retyme, doer)  where:
            dog is generator
            retyme is tyme (real or simulated) in seconds when dog should run next
            doer is from .doers list used to assign its .done state given
                completion state of its dog

        Calls each generator callable (instance or function or method) in .doers
        to create each generator dog. Injects own tymth function closure, and
            generator function's own tock, and opts.

        Runs enter context of each dog by calling next(dog)

        Parameters:
            doers is list of generator method or function callables with attributes
                .tock is tyme increment in seconds
                .done is Boolean completion state
                .opts is dict() of optional parameters
                If not provided uses .doers.
                The normal case is to initialize in .__init__. or .do().
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any

        Returns:
            deeds deque():
                A deed is tuple of form (dog, retyme, doer). If not provided
                uses .deeds.

        See: https://stackoverflow.com/questions/40528867/setting-attributes-on-func
        For setting attributes on bound methods.
        """

        if doers is None:
            doers = self.doers
            deeds = self.deeds
        else:
            deeds = deque()  # when doers is provided then don't use .deeds

        for doer in doers:
            try:
                doer.done = False  # False at enter. False signals incomplete
            except AttributeError:  # when using bound method for generator function
                doer.__func__.done = False  # False at enter. False signals incomplete

            temp = temp or (doer.temp if hasattr(doer, "temp") and doer.temp else None)
            opts = doer.opts if hasattr(doer, "opts") else {}

            dog = doer(tymth=self.tymen(), tock=doer.tock, temp=temp, **opts)  # calls doer.do
            try:
                next(dog)  # run enter by advancing to first yield
            except StopIteration as ex:   # return not yield
                # done in enter so assign done state
                try:  # assign done state non forced return
                    doer.done = ex.value if ex.value is not None else doer.done
                except AttributeError:  # bount method generator
                    # write to doer.__func__.done read from doer.done
                    doer.__func__.done = ex.value if ex.value is not None else doer.done
                #try:
                    #doer.done = ex.value if ex.value else False  # assign done state
                #except AttributeError:
                    #doer.__func__.done = ex.value if ex.value else False  # assign done state
                continue  # don't append
            deeds.append((dog, self.tyme, doer))
        return deeds


    def recur(self, deeds=None):
        """
        Recur once through deeds deque of tuples (triples) of form
        (dog, retyme, doer) and update in place

        Each deed is deque of tuples of  form (dog, retyme, doer) where:
            dog is generator
            retyme is tyme (real or simulated) in seconds when dog should run next
            doer is from .doers list used to assign its .done state given
            associated completion state of its  dog

        Each cycle checks all generators in deeds deque and runs if retyme past.
        At end of cycle advances .tyme by one .tock by calling .tick()

        Parameters:
            deeds (deque):  tuples of form (dog, retyme, doer).
                    Parameterization here of deeds enables some special cases.

        The Parameterization here of deeds enables some special cases
        such as manual testing or iteraton.
        The normal case is to initialize .doers in .__init__. or .do() and to
        initialize .deeds in .__init__. and then update in .enter()
        """
        if deeds is None:
            deeds = self.deeds

        deeds.append((None, None, None))  # append run through once marker
        while deeds: # do while uses explicit break to exit while
            dog, retyme, doer = deeds.popleft()  # pop it off
            if not dog:  # Marker detected so this run through once has completed
                break  # break loop at marker signifies once through

            if retyme <= self.tyme:  # run it now
                try:  # send tyme. yield tock, tock may change during sended run
                    tock = dog.send(self.tyme)  # yielded tock == 0.0 means re-run asap
                except StopIteration as ex:  # returned instead of yielded
                    try:  # assign done state non forced return
                        doer.done = ex.value if ex.value is not None else doer.done
                    except AttributeError:  # bount method generator
                        # write to doer.__func__.done read from doer.done
                        doer.__func__.done = ex.value if ex.value is not None else doer.done
                else:  # reappend for next pass
                    if not tock:  # tock is None or tock == 0.0 with empty yield tock == None
                        retyme = self.tyme + self.tock  # rerun at next recur
                    else:
                        retyme += tock  # cumulative retyme of doer tock
                    deeds.append((dog, retyme, doer))  # reappend for next run through
            else:  # not retyme yet
                deeds.append((dog, retyme, doer))  # reappend for next pass

        self.tick()  # advance .tyme by one doist .tock


    def exit(self, deeds=None):
        """
        Force exit each still opened deed calling .close on the dog generator
        which throws a GeneratorExit to the generator.
        This executes the close context (GeneratorExit) which then excecutes
        the exit context in the finally caluse. Each dogs exit is responsible
        for releasing resources
        Previously aborted or closed dogs have already exited
        Close any running dogs in reverse order so that enters and exits are
        nested pairs so that the corresponding exits appear in reverse order
        to their entes. This preserves nested resource dependencies.
        For example:
            enter A,
                enter B,
                    enter C,
                    exit C,
                exit B,
            exit A

        Parameters:
            deeds (deque): tuples of form (dog, retyme, doer).
                If not provided uses .deeds.
                Parameterization here of deeds enables some special cases.
        """
        if deeds is None:
            deeds = self.deeds

        while(deeds):  # .close each remaining dog in deeds in reverse order
            dog, retime, doer = deeds.pop()  # pop it off in reverse (right side)
            if not dog:  # marker deed
                continue  # skip marker
            try:
                done = dog.close()  # force GeneratorExit. Maybe log exit tock tyme
            except StopIteration:
                pass  # Hmm? Not supposed to happen!
            else:  # set done state forced close
                try:  # not bound method generator but doer instance or function
                    doer.done = done if done is not None else doer.done
                except AttributeError:  # when using bound method generator
                    # writing to doer.__func__.done read from doer.done
                    doer.__func__.done = done if done is not None else doer.done


    def extend(self, doers):
        """
        Extend .doers list with doers. Ready deeds from doers and extend .doers
        and .deeds.  Edit deeds in place so not replace deque.

        Parameters:
            doers is list of doers to add as extension.

        """
        doers = [doer for doer in doers if doer not in self.doers] # ensure unique
        deeds = self.enter(doers=doers)  # provide fresh deeds for new doers
        self.doers.extend(doers)
        self.deeds.extend(deeds)


    def remove(self, doers):
        """
        Remove doers from .doers list and any associated deeds from .deeds deque.
        Force close removed deeds.

        Parameters:
            doers is list of doers to remove.

        """
        rdoers = [doer for doer in doers if doer in self.doers] # ensure in .doers
        rdeeds = deque()  # fresh deque for deeds to remove
        deeds = self.deeds  # edit update self.deeds in place
        for i in range(len(deeds)):  # iterate once over each deed
            dog, retyme, doer = deeds.popleft()
            if not dog:  # reappend the run through once marker deed
                deeds.append((dog, retyme, doer))
            elif doer in rdoers:  # found deed to remove and close
                rdeeds.append((dog, retyme, doer))  # add to removal deque
            else:  # keep deed do not remove and close
                deeds.append((dog, retyme, doer))  # reappend

        for doer in rdoers:  # update .doers to remove rdoers
            self.doers.remove(doer)

        self.exit(deeds=rdeeds)


def doify(f, *, name=None, tock=0.0, temp=None, **opts):
    """Returns Doist/DoDoer compatible copy, g, of converted generator
    function/method f.
    Each doify(f) invoction returns a unique copy of doified function/method f.
    Imbues copy, g, of converted generator function/method, f, with attributes
    used by Doist.enter() or DoDoer.enter().
    Allows multiple instances of copy, g, of generator function/method, f, each with
    unique attributes.

    Usage:
    def f():
       pass

    c = doify(f, name='c')

    Parameters:
        f (function): generator function
        name (str): new function name for returned doified copy g. Default is to copy
            f.__name__
        tock (float): default tock attribute of doified copy g
        temp (bool | None): use temporay file resources if any
        opts (dict): remaining parameters that becomes .opts attribute
            of doified copy g

    Based on:
    https://stackoverflow.com/questions/972/adding-a-method-to-an-existing-object-instance
    """
    g = helping.copyfunc(f, name=name)
    g.done = None  # default done state
    g.tock = tock  # default tock attribute
    g.temp = temp  # default temp attribute
    g.opts = dict(opts)  #  default opts attribute
    if inspect.ismethod(f):  # f.__self__ instance method
        g = types.MethodType(g, f.__self__)  # make g a method of f.__self__ only
    return g


def doize(*, tock=0.0, temp=None, **opts):
    """Returns decorator that makes decorated generator function Doist compatible.
    Imbues decorated generator function with attributes used by Doist.enter() or
    DoDoer.enter().
    Only one instance of decorated function with shared attributes is allowed.

    Usage:
    @doize
    def f():
       pass

    Parameters:
        tock (float): default tock attribute of doized f
        opts (dict): remaining parameters that becomes .opts attribute
            of doized f
    """
    def decorator(f):
        f.done = None  # default done state
        f.tock = tock  # default tock attribute
        f.temp = temp  # default temp attribute
        f.opts = dict(opts)  # default opts attribute
        return f
    return decorator


class Doer(tyming.Tymee):
    """
    Doer base class for hierarchical structured async coroutine like generators.
    Doer.__call__ on instance returns generator.
    Interface for Doist etc is generator function like object.
    Doer is generator method instance creator and has extra methods and
    attributes that a plain generator function does not

    The .do method executes other methods each corresponding to one of the
    six econtexts:
        enter, recur, clean, exit, cease (forced), abort (forced)

    Actual context order may be one of:
        enter, recur, clean, exit
        enter, recur, cease, exit
        enter, recur, abort, exit
        enter, abort, exit

    Attributes:
        done (bool): completion state:
            True means completed
            Otherwise incomplete. Incompletion maybe due to close or abort.
        opts (dict): injected options into its .do generator by scheduler
        temp (bool): True means use temporary file resources if any

    Inherited Properties:
        tyme (float): is float relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        tymth (closure): function wrapper closure returned by Tymist.tymen()
                        method. When .tymth is called it returns associated
                        Tymist.tyme. Provides injected dependency on Tymist
                        tyme base.

    Properties:
        tock (float): desired time in seconds between runs or until next run,
                 non negative, zero means run asap

    Inherited Methods:
        wind  injects ._tymth dependency from associated Tymist to get its .tyme

    Methods:
        .__call__ makes instance callable
            Appears as generator function that returns generator
        .do is generator method that returns generator
        .enter is enter context action method
        .recur is recur context action method or generator method
        .clean is clean context action method
        .exit is exit context method
        .cease is cease context method
        .abort is abort context method

    Hidden:
        _tymth (closure): is injected function wrapper closure returned by .tymen() of
            associated Tymist instance that returns Tymist .tyme. when called.
        _tock (float): is hidden attribute for .tock property

    """

    def __init__(self, *, tymth=None, tock=0.0, opts=None, temp=False, **kwa):
        """
        Initialize instance.

        Inherited Parameters:
            tymth (closure): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.

        Parameters:
           tock (float): seconds initial value of .tock
           opts (dict): injected options into its .do generator by scheduler
           temp (bool): True means use temporary file resources if any

        """
        super(Doer, self).__init__(tymth=tymth, **kwa)
        self.done = None  #  default completion state
        self.tock = tock  # desired tyme interval between runs, 0.0 means asap
        # used for injection of options into .do by scheduler
        self.opts = opts if opts is not None else {}  # empty dict if None
        self.temp = True if temp else False


    def __call__(self, **kwa):
        """
        Returns generator
        Does not advance to first yield.
        The advance to first yield effectively invokes the enter or open context
        on the generator.
        To enter either call .next or .send(None) on generator
        """
        return self.do(**kwa)


    @property
    def tock(self):
        """
        tock property getter, get ._tock
        .tock is float desired .tyme increment in seconds
        """
        return self._tock


    @tock.setter
    def tock(self, tock):
        """
        desired cycle tyme interval until next run
        0.0 means run asap,
        set ._tock to tock
        """
        self._tock = abs(float(tock))


    def do(self, tymth, *, tock=0.0, temp=None, **opts):
        """
        Generator method to run this doer.
        Calling this method returns generator.
        Interface matches generator function for compatibility.
        To customize create subclasses and override the lifecycle methods:
            .enter, .recur, .exit, .cease, .abort

        Parameters:
            tymth (closure): function wrapper closure returned by Tymist.tymen()
                        method. When .tymth is called it returns associated
                        Tymist.tyme. Provides injected dependency on Tymist
                        tyme base.
            tock (float): injected initial tock value
            temp (bool): True means use temporary file resources if any
            opts (dict): of injected optional additional parameters
        """
        temp = temp or (self.temp if self.temp else temp)  # inject if temp or self.temp

        try:
            # enter context
            self.wind(tymth)  # update tymist dependencies
            self.tock = tock  #  set tock to parameter
            self.enter(temp=temp)

            #recur context self.done set by super doist or dodoer prior to enter
            if isgeneratorfunction(self.recur):  #  .recur is generator method
                self.done = yield from self.recur()  # recur context delegated
            else:  # .recur is standard method so iterate in while loop
                while (not self.done):  # recur context
                    tyme = (yield (self.tock))  # yields .tock then waits for next send
                    self.done = self.recur(tyme=tyme)  # False means recur again

        except GeneratorExit:  # close context, forced exit due to .close on generator
            self.cease()

        except Exception as ex:  # abort context, forced exit due to uncaught exception
            self.abort(ex=ex)
            raise

        else:  # clean context
            self.clean()

        finally:  # exit context, exit, unforced if normal exit of try, forced otherwise
            self.exit()

        # return value of yield from or StopIteration.value indicates completion
        # python 3.13 gh-104770: If a generator returns a value upon being
        # closed, the value is now returned by generator.close().
        return self.done  # Only returns done state if normal return or close not abort raise


    def enter(self, *, temp=None):
        """Do 'enter' context actions. Override in subclass. Not a generator method.
        Set up resources. Comparable to context manager enter.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any
        """
        # inject temp into file resources here if any


    def recur(self, tyme):
        """
        Do 'recur' context actions. Override in subclass.
        Regular method that perform repetitive actions once per invocation.
        Assumes resource setup in .enter() and resource takedown in .exit()
        (see ReDoer below for example of .recur that is a generator method)

        Returns completion state of recurrence actions.
           True means done False means continue

        Parameters:
            Doist feeds its .tyme through .send to .do yield which passes it here.


        .recur maybe implemented by a subclass either as a non-generator method
        or a generator method. This stub here is as a non-generator method.
        The base class .do detects which type:
            If non-generator .do method runs .recur method once per iteration
                until .recur returns (True)
            If generator .do method runs .recur with (yield from) until .recur
                returns (see ReDoer for example of generator .recur)

        """
        return (False)


    def clean(self):
        """
        Do 'clean' context actions. Override in subclass. Not a generator method.
        Clean up resources that are unique to a clean exit.
        Called by else after normal return.
        """


    def exit(self):
        """
        Do 'exit' context actions. Override in subclass. Not a generator method.
        Clean up resources. Comparable to context manager exit.
        Called by finally after normal return, close, or abort.
        After .exit() do returns resulting in StopIteration.
        """


    def cease(self):
        """
        Do 'cease' context actions. Forced exist. Override in subclass.
        Not a generator method.
        Forced cease by thrown generator.close() method causing GeneratorExit.
        .exit() is finally called after .cease().
        """


    def abort(self, ex):
        """
        Do 'abort' context actions. Override in subclass. Not a generator method.
        Parameters:
            ex is Exception instance that caused abort.
        Unexpected exception that results in generator exiting but not GeneratorExit.
        .exit() is finally called after .abort().
        """


class ReDoer(Doer):
    """
    ReDoer is an example sub class whose .recur is a generator method not a
    plain method. Its .do method detects that its .recur is a generator method
    and executes it using yield from instead of just calling the method.

    Inherited Attributes:
        done (bool | None): completion state: True means completed
            Otherwise incomplete. Incompletion maybe due to close or abort.
        opts (dict): injected options into its .do generator by scheduler
        temp (bool | None): use temporary file resources if any

    Inherited Properties:
        tyme (float): relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        tymth (closure): function wrapper closure returned by Tymist.tymen() method.
            When .tymth is called it returns associated Tymist.tyme.
            .tymth provides injected dependency on Tymist tyme base.
        tock (float): desired time in seconds between runs or until next run,
                 non negative, zero means run asap

    Inherited Methods:
        .wind  injects ._tymth dependency from associated Tymist to get its .tyme
        .__call__ makes instance callable
            Appears as generator function that returns generator
        .do is generator method that returns generator
        .enter is enter context action method
        .recur is recur context action method or generator method
        .exit is exit context method
        .close is close context method
        .abort is abort context method

    Overidden Methods:
        .recur

    Hidden:
       ._tymth is injected function wrapper closure returned by .tymen() of
            associated Tymist instance that returns Tymist .tyme. when called.
       ._tock is hidden attribute for .tock property

    """

    def recur(self):
        """
        Do 'recur' context actions as a generator method. Override in subclass.
        Assumes resource setup in .enter() and resource takedown in .exit()
        (see Doer for example of .recur that is a regular method)

        yield the current .tock
        accepts the current tyme
        return value is used for .done  (true done false not done but ended)

        Parameters:
            tyme is initial output of send fed to do yield, Doist feeds its .tyme
        Returns completion state of recurrence actions.
           True means done False means continue

        For base class do:
            yield from this generator recur method which runs until returns

        """
        count = 0
        # print("ReDoer recur before yield. tyme = {} count={}\n".format(tyme, count))
        while (True):  # recur context
            tyme = yield(self.tock)  # first yield of None
            count += 1
            # print("ReDoer recur after yield. tyme = {} count={}\n".format(tyme, count))
            if count >= 3:
                break

        # print("ReDoer recur after break tyme = {} count={}\n".format(tyme, count))
        return True  # done



class DoDoer(Doer):
    """
    DoDoer implements Doist like functionality to allow nested scheduling of Doers.
    Each DoDoer runs a list of doers like a Doist but using the tyme from its
       injected tymth for the associated tymist as injected by its ultimate root
       parent Doist and any intervening parent DoDoer(s).

    Scheduling hierarchy: Doist->DoDoer...->DoDoer->Doers

    Inherited Attributes:
        done (bool): completion state:
                    True means completed
                    Otherwise incomplete. Incompletion maybe due to close or abort.
        opts (dict): injected options for its generator .do
        temp (bool | None): use temporary file resources if any

    Attributes:

    Inherited Properties:
        .tyme is float relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        .tymth is function wrapper closure returned by Tymist .tymeth() method.
            When .tymth is called it returns associated Tymist .tyme.
            .tymth provides injected dependency on Tymist tyme base.
        .tock is float, desired time in seconds between runs or until next run,
                 non negative, zero means run asap

    Properties:
        doers (list): Doer or Doist compatible generator instances,
            functions, or methods.
        deeds (deque):  tuples of form (dog, retyme, doer)  where:
            dog is generator created by doer.
            retyme is tyme in seconds when next should run may be real or simulated.
            doer is associated doer in .doers list.
            Used throughout the execution lifecycle. The normal
            case is use the default empty initialization performed here and
            update in .enter().
        always (bool): True means keep running even when all dogs in deeds
            are complete. Enables dynamically managing extending or removing
            doers and associated deeds while running.

    Inherited Methods:
        .wind  injects ._tymth dependency from associated Tymist to get its .tyme
        .__call__ makes instance callable
            Appears as generator function that returns generator
        .do is generator method that returns generator
        .enter is enter context action method
        .recur is recur context action method or generator method
        .clean is clean context action method
        .exit is exit context method
        .close is close context method
        .abort is abort context method

    Overidden Methods:
        .do
        .enter
        .recur
        .exit

    Hidden:
       ._tymth is injected function wrapper closure returned by .tymen() of
            associated Tymist instance that returns Tymist .tyme. when called.
       ._tock is hidden attribute for .tock property
       ._always is hidden attribute for .always property
       ._doers is hidden attribute for .doers property
       ._deeds is hidden attribute for .deeds property

    """

    def __init__(self, doers=None, always=False, **kwa):
        """
        Initialize instance.

        Inherited Parameters:
            tymth is injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock is float seconds initial value of .tock

        Parameters:
            doers (iterable): Doer class instances, generator methods or function
                callables with attributes tock, done, and opts dict() used to
                initialize .doers.
                The .doers attribute is used throughout the execution lifecycle.
                Parameterization elsewhere of doers enables some special cases.
                The normal case is to initialize here.

            always is Boolean, True means keep running even when all dogs in deeds
                are complete. Enables dynamically managing extending or removing
                doers and associated deeds while running.

        """
        super(DoDoer, self).__init__(**kwa)
        self.doers = list(doers) if doers is not None else []
        self.deeds = deque()
        self.always = always


    @property
    def doers(self):
        """
        doers property getter, get ._doers
        .doers is list of doist compatible generator instances, functions, or
            methods
        """
        return self._doers


    @doers.setter
    def doers(self, doers):
        """
        set ._doers to doers list
        """
        if not isinstance(doers, list):
            raise TypeError("Expected list, got {}.".format(type(doers)))
        self._doers = doers


    @property
    def deeds(self):
        """
        deeds property getter, get ._deeds
        .deeds is deque of triples, each of form (dog, retyme, doer)
        """
        return self._deeds


    @deeds.setter
    def deeds(self, deeds):
        """
        set ._deeds to deeds deque
        """
        if not isinstance(deeds, deque):
            raise TypeError("Expected deque, got {}.".format(type(deeds)))
        self._deeds = deeds


    @property
    def always(self):
        """
        always property getter, get ._always
        .always is Boolean, True means keep running even when all dogs in deeds
            are complete. Enables dynamically managing extending or removing
            doers and associated deeds while running.
        """
        return self._always


    @always.setter
    def always(self, always):
        """
        set ._always to always
        """
        self._always = True if always else False


    def do(self, tymth, tock=0.0, doers=None, always=None, *, temp=None, **opts):
        """
        Generator method to run this doer. Equivalent of doist.do
        Calling this method returns generator
        Interface matched generator function for compatibility

        Parameters:
            tymth (closure): injected function wrapper closure returned by
                Tymist.tymen(). Calling tymth() returns associated Tymist.tyme.
            tock (float): injected initial tock value
            doers (list): of generator method or function callables with attributes
                tock, done, and opts dict(). This may be used to update the .doers
                attribute which is used throughout the execution lifecycle.
                If not provided uses .doers.
                Parameterization here of doers enables some special cases.
                The normal case is to initialize in .__init__.
            always (bool): True means keep running even when all dogs in deeds
                are complete. Enables dynamically managing extending or removing
                doers and associated deeds while running.
                When not provided use .always.
                temp (bool): True means use temporary file resources if any
            opts (dict): injected optional additional parameters
        """
        temp = temp or (self.temp if self.temp else temp)  # inject if temp or self.temp

        always = always if always is not None else self.always
        if doers is not None:
            self.doers = list(doers)
            self.deeds = deque()

        try:
            # enter context
            self.wind(tymth)  # change tymist dependencies
            self.tock = tock  #  set tock to parameter
            # tyme = self.tyme

            self.enter(temp=temp)  # doist.enter() equivalent

            #recur context  self.done set by super doist or dodoer prior to enter
            while (not self.done or always):  # recur context
                tyme = (yield (self.tock))  # yields .tock then waits for next send
                self.done = self.recur(tyme=tyme)  # equv of doist.recur

        except GeneratorExit:  # cease context, forced exit due to generator.close()
            self.cease()

        except Exception as ex:  # abort context, forced exit due to uncaught exception
            self.abort(ex=ex)
            raise

        else:  # clean context
            self.clean()

        finally:  # exit context, exit, unforced if normal exit of try, forced otherwise
            self.exit()  # equiv of doist.do finally clause

        # return value of yield from or StopIteration.value indicates completion
        return self.done  # Only returns done state if normal return not close or abort raise


    def enter(self, doers=None, *, temp=None):
        """Do 'enter' context actions. Equivalent of Doist.enter()
        Set up resources. Comparable to context manager enter.

        Returns deeds deque of triples (dog, retyme, doer)  where:
            dog is generator created by doer
            retyme is tyme in seconds when next should run may be real or simulated
            doer is doer for dog from doers list

        Calls each generator callable (function or method) in .doers to create
        each generator dog.

        Runs enter context of each dog by calling next(dog)

        Parameters:
            doers (list): Doer Instance, generator method or function callables
                with attributes tock, done, and opts dict().
                If not provided uses .doers.
                Parameterization here of doers enables some special cases.
                The normal case is to initialize in .__init__.

            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any

        Returns:
            deeds (deque): A deed is tuple of form (dog, retyme, doer).
                           If not provided uses .deeds.

        See: https://stackoverflow.com/questions/40528867/setting-attributes-on-func
        For setting attributes on bound methods.
        """
        # inject temp into file resources here if any

        if doers is None:
            doers = self.doers
            deeds = self.deeds
        else:
            deeds = deque()

        for doer in doers:
            try:
                doer.done = False  # False at enter. False signals incomplete
            except AttributeError:   # when using bound method for generator function
                doer.__func__.done = False  # False at enter.  False signals incomplete
            temp = temp or (doer.temp if hasattr(doer, "temp") and doer.temp else None)
            opts = doer.opts if hasattr(doer, "opts") else {}

            dog = doer(tymth=self.tymth, tock=doer.tock, temp=temp, **opts)  # calls doer.do
            try:
                next(dog)  # run enter by advancing to first yield
            except StopIteration as ex:  # return not yield
                # done in enter so assign done state
                try:  # assign done state non forced return
                    doer.done = ex.value if ex.value is not None else doer.done
                except AttributeError:  # bount method generator
                    # write to doer.__func__.done read from doer.done
                    doer.__func__.done = ex.value if ex.value is not None else doer.done
                #try:
                    #doer.done = ex.value if ex.value else False  # assign done state
                #except AttributeError:
                    #doer.__func__.done = ex.value if ex.value else False  # assign done state


                continue  # don't append already complete
            deeds.append((dog, self.tyme, doer))
        return deeds


    def recur(self, tyme, deeds = None):
        """
        Do 'recur' context actions. Equivalent of Doist.recur

        Parameters:
            tyme (float): is output of send fed to do yield, The root scheduler
                Doist feeds its .tyme which propagates down the chain of DoDoers
                Because tymist is injected by doist or dodoer, self.tyme is same
                as tyme. So may use either which is more convenient.
             deeds (deque): tuples of form (dog, retyme, doer).
                If not provided uses .deeds.
                Parameterization here of deeds enables some special cases.

        Returns completion state of recurrence actions.
           True means done False means continue

        Cycle once through deeds deque and update in place

        Each cycle checks all generators dogs in deeds deque and runs if retyme past.
        """
        if deeds is None:
            deeds = self.deeds

        deeds.append((None, None, None))  # append run through once marker
        while deeds:  # do while uses explicit break to exit while
            dog, retyme, doer = deeds.popleft()  # pop it off
            if not dog:  # Marker detected so this run through once has completed
                break  # break loop at marker signifies once through

            if retyme <= tyme:  # run it now
                try:  # send tyme. yield tock, tock may change during sended run
                    tock = dog.send(tyme)  # yielded tock == 0.0 means re-run asap
                except StopIteration as ex:  # returned instead of yielded
                    try:  # assign done state non forced return
                        doer.done = ex.value if ex.value is not None else doer.done
                    except AttributeError:  # bount method generator
                        # write to doer.__func__.done read from doer.done
                        doer.__func__.done = ex.value if ex.value is not None else doer.done
                else:  # reappend for next pass
                    if not tock:  # tock is None or tock == 0.0 with empty yield tock == None
                        retyme = tyme + self.tock  # rerun at next recur
                    else:
                        retyme += tock  # cumulative retyme of doer tock
                    deeds.append((dog, retyme, doer))  # reappend for next run through
            else:  # not retyme yet
                deeds.append((dog, retyme, doer))  # reappend for next run through

        return (not deeds)  # True if deeds deque is empty


    def exit(self, deeds = None):
        """
        Do 'exit' context actions.

        Parameters:
            deeds (deque): of deed tuples of form (dog, retyme, doer)
                If not provided uses .deeds.
                Parameterization here of deeds enables some special cases.

        See: https://stackoverflow.com/questions/40528867/setting-attributes-on-func
        For setting attributes on bound methods.
        """
        if deeds is None:
            deeds = self.deeds

        while(deeds):  # .close each remaining dog in deeds in reverse order
            dog, retime, doer = deeds.pop()  # pop it off in reverse (right side)
            if not dog:  # marker deed
                continue  # skip marker
            try:
                done = dog.close()  # force GeneratorExit returns None if already closed or Gen Return value
            except StopIteration:
                pass  # Hmm? Not supposed to happen!
            else:  # set done state forced exit
                try:  # not bound method generator but doer instance or function
                    doer.done = done if done is not None else doer.done
                except AttributeError:  # when using bound method generator
                    # writing to doer.__func__.done read from doer.done
                    doer.__func__.done = done if done is not None else doer.done


    def extend(self, doers):
        """
        Extend .doers list with doers. Ready deeds from doers and extend .doers
        and .deeds.  Edit deeds in place so not replace deque.

        Parameters:
            doers is list of doers to add as extension.

        """
        doers = [doer for doer in doers if doer not in self.doers] # ensure unique
        deeds = self.enter(doers=doers)  # provide fresh deeds for new doers
        self.doers.extend(doers)
        self.deeds.extend(deeds)


    def remove(self, doers):
        """
        Remove doers from .doers list and any associated deeds from .deeds deque.
        Force close removed deeds.

        Parameters:
            doers is list of doers to remove.

        """
        rdoers = [doer for doer in doers if doer in self.doers] # ensure in .doers
        rdeeds = deque()  # fresh deque for deeds to remove
        deeds = self.deeds  # edit update self.deeds in place
        for i in range(len(deeds)):  # iterate once over each deed
            dog, retyme, doer = deeds.popleft()
            if not dog:  # reappend the run through once marker deed
                deeds.append((dog, retyme, doer))
            elif doer in rdoers:  # found deed to remove and close
                rdeeds.append((dog, retyme, doer))  # add to removal deque
            else:  # keep deed do not remove and close
                deeds.append((dog, retyme, doer))  # reappend

        for doer in rdoers:  # update .doers to remove rdoers
            self.doers.remove(doer)

        self.exit(deeds=rdeeds)


def bareDo(tymth=None, tock=0.0, *, temp=None, **opts):
    """
    Bare bones generator function template as example of generator function
    suitable for use with either doify wrapper or doize decorator.
    Make copy and rename for given application.
    Calling copied renamed function returns basic generator.
    Wrapping copied renamed function with doify returns yet unique wrapped copy
    with unique values of injected attributes and  parameters and further
    renamed by wrapper.
    Decorating copied renamed function with doize returns singleton with injected
    parameter values.

    Injected Attributes:
        g.tock = tock  # default tock attributes
        g.done = None  # default done state
        g.temp = None  # temporary resources
        g.opts = opts

    Parameters:
        tymth (closure): injected function wrapper closure returned by
                Tymist.tymen(). Calling tymth() returns associated Tymist.tyme.
        tock (float): injected initial tock value
        temp (bool): True means use temporary file resources if any
        opts (dict):  injected optional additional parameters

    The function comments show where the 6 equivalent contexts are performed
    enter, recur, clean, exit, (unforced) close, abort (forced)
    So context order may be:
    enter, recur, clean, exit
    enter, recur, cease, exit
    enter, recur, abort, exit
    enter, abort, exit
    """
    try:
        # enter context
        # tyme = tymth()
        done = False

        while not done:  # recur context
            tyme = (yield (tock))  # yields tock then waits for next send
            #  do stuff repeately in while loop
            done = True  # means ready to exit while loop

    except GeneratorExit:  # cease context upon Doist thrown .close to force early exit.
        pass  # do forced close clean up here

    except Exception as ex:  # abort context, forced exit due to uncaught exception
        pass  # do unexpected exception clean up here
        raise ex  # always re-raise exception after any exception specific actions

    else:  # clean context, clean exit processing when not (close or abort)
        pass  # do clean exit only clean up here

    finally:  # exit context,  all exits both forced and unforced come through here
        pass  # manditory clean up here

    return (done)  # returned value on final yield from, or yielded ex.value of StopIteration


class ExDoer(Doer):
    """
    ExDoer is example Doer for testing and demonstration
    Supports introspection with methods to record sends and yields

    See Doer for inherited attributes, properties, and methods.

    Attributes:
       states (list): State namedtuples (tyme, context, feed, count)
       count (int): iteration count

    """

    def __init__(self, **kwa):
        """
        Initialize instance.
        """
        super(ExDoer, self).__init__(**kwa)
        self.states = []
        self.count = None


    def enter(self, *, temp=None):
        """Do 'enter' context actions. Override in subclass. Not a generator method.
        Set up resources. Comparable to context manager enter.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any

        Doist or DoDoer winds its doers on enter
        """
        # inject temp into file resources here if any
        self.count = 0
        self.states.append(State(tyme=self.tyme, context="enter",
                                 feed=self.tyme, count=self.count))

    def recur(self, tyme):
        """"""
        self.count += 1
        self.states.append(State(tyme=self.tyme, context="recur",
                                 feed=self.tyme, count=self.count))
        if self.count > 3:
            return True  # complete
        return False  # incomplete

    def exit(self):
        """"""
        self.count += 1
        self.states.append(State(tyme=self.tyme, context='exit',
                                 feed=None, count=self.count))

    def cease(self):
        """"""
        self.count += 1
        self.states.append(State(tyme=self.tyme, context='cease',
                                 feed=None, count=self.count))

    def abort(self, ex):
        """"""
        self.count += 1
        self.states.append(State(tyme=self.tyme, context='abort',
                                 feed=ex.args[0], count=self.count))




def doifyExDo(tymth, tock=0.0, states=None, *, temp=None, **opts):
    """
    Example generator function for testing and demonstration.
    Example non-class based generator for use with doify wrapper.
    Calling this function returns generator.
    Wrapping this function with doify returns copy with unique attributes

    Parameters:
        tymth (closure): injected function wrapper closure returned by
                Tymist.tymen(). Calling tymth() returns associated Tymist.tyme.
        tock (float): injected initial tock value
        states (list): State namedtuples (tyme, context, feed, count)
        temp (bool): True means use temporary file resources if any
        opts (dict): injected optional additional parameters

    """
    tyme = tymth()
    count = 0
    states = states if states is not None else []

    try:
        # enter context
        states.append(State(tyme=tymth(), context="enter", feed=tyme, count=count))
        while (True):  # recur context
            tyme = (yield (tock))  # yields tock then waits for next send
            count += 1
            states.append(State(tyme=tymth(), context="recur", feed=tyme, count=count))
            if count > 3:
                break  # normal exit

    except GeneratorExit:  # cease context, forced exit due to generator.close
        count += 1
        states.append(State(tyme=tymth(), context='cease', feed=None, count=count))

    except Exception as ex:  # abort context, forced exit due to uncaught exception
        count += 1
        states.append(State(tyme=tymth(), context='abort', feed=ex.args[0], count=count))
        raise ex

    finally:  # exit context,  unforced exit due to normal exit of try
        count += 1
        states.append(State(tyme=tymth(), context='exit', feed=None, count=count))

    return (True)  # return value of yield from, or yield ex.value of StopIteration


@doize(tock=0, states=None)
def doizeExDo(tymth, tock=0.0, states=None, *, temp=None, **opts):
    """
    Example decorated generator function for use with doize decorator.
    Example non-class based generator
    Calling this function returns generator

    Parameters:
        tymth (closure): injected function wrapper closure returned by
                Tymist.tymen(). Calling tymth() returns associated Tymist.tyme.
        tock (float): injected initial tock value
        states (list): of State namedtuples (tyme, context, feed, count)
        temp (bool): True means use temporary file resources if any
        opts (dict): injected optional additional parameters
    """
    tyme = tymth()
    count = 0
    states = states if states is not None else []

    try:
        # enter context
        states.append(State(tyme=tymth(), context="enter", feed=tyme, count=count))
        while (True):  # recur context
            tyme = (yield (tock))  # yields tock then waits for next send
            count += 1
            states.append(State(tyme=tymth(), context="recur", feed=tyme, count=count))
            if count > 3:
                break  # normal exit

    except GeneratorExit:  # cease context, forced exit due to generator.close
        count += 1
        states.append(State(tyme=tymth(), context='cease', feed=None, count=count))

    except Exception as ex:  # abort context, forced exit due to uncaught exception
        count += 1
        states.append(State(tyme=tymth(), context='abort', feed=ex.args[0], count=count))
        raise ex

    finally:  # exit context,  unforced exit due to normal exit of try
        count += 1
        states.append(State(tyme=tymth(), context='exit', feed=None, count=count))

    return (True)  # return value of yield from, or yield ex.value of StopIteration




#  for testing purposes
class TryDoer(Doer):
    """
    TryDoer supports testing with methods to record sends and yields

    Inherited Attributes:
        done (bool | None): completion state:
                             True means completed Otherwise incomplete.
                             Incompletion maybe due to close or abort.
        opts (dict): injected options into its .do generator by scheduler
        temp (bool | None): use temporary file resources if any

    Attributes:
       states (list): State namedtuples (tyme, context, feed, count)
       count (int): is context count
       stop (int): count where doer completes

    Inherited Properties:
        tyme (float): relative cycle time of associated Tymist .tyme obtained
            via injected .tymth function wrapper closure.
        tymth (closure): injected function wrapper closure returned by
                Tymist.tymen(). Calling tymth() returns associated Tymist.tyme.
        tock (float): injected initial tock value. non negative, zero means run asap


    Properties:

    Methods:
        .wind  injects ._tymth dependency from associated Tymist to get its .tyme
        .__call__ makes instance callable
            Appears as generator function that returns generator
        .do is generator method that returns generator
        .enter is enter context action method
        .recur is recur context action method or generator method
        .exit is exit context method
        .close is close context method
        .abort is abort context method
    """

    def __init__(self, stop=3, **kwa):
        """
        Initialize instance.
        Inherited Parameters:
           tymist is Tymist instance
           tock is float seconds initial value of .tock
        Parameters:
           stop is count when complete

        """
        super(TryDoer, self).__init__(**kwa)
        self.states = []
        self.count = None
        self.stop = stop

    def enter(self, *, temp=None):
        """Do 'enter' context actions. Override in subclass. Not a generator method.
        Set up resources. Comparable to context manager enter.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any
        """
        # inject temp into file resources here if any

        feed = "Default"
        self.count = 0
        self.states.append(State(tyme=self.tyme, context="enter",
                                 feed=feed, count=self.count))

    def recur(self, tyme):
        """

        """
        self.count += 1
        self.states.append(State(tyme=self.tyme, context="recur",
                                 feed=tyme, count=self.count))
        if self.count > self.stop:
            return True  # complete
        return False  # incomplete

    def exit(self):
        """
        """
        self.count += 1
        self.states.append(State(tyme=self.tyme, context='exit',
                                 feed=None, count=self.count))

    def cease(self):
        """
        """
        self.count += 1
        self.states.append(State(tyme=self.tyme, context='cease',
                                 feed=None, count=self.count))

    def abort(self, ex):
        """
        """
        self.count += 1
        self.states.append(State(tyme=self.tyme, context='abort',
                                 feed=ex.args[0], count=self.count))


#  for testing purposes
@doize()
def tryDo(states, tymth, tock=0.0, *, temp=None, **opts):
    """Generator function test example non-class based generator.
    Calling this function returns generator:

    Parameters:
        tymth (closure): injected function wrapper closure returned by
                Tymist.tymen(). Calling tymth() returns associated Tymist.tyme.
        tock (float): injected initial tock value
        temp (bool): True means use temporary file resources if any
        opts (dict): injected optional additional parameters

    """
    feed = "Default"
    count = 0

    try:
        # enter context

        states.append(State(tyme=tymth(), context="enter", feed=feed, count=count))
        while (True):  # recur context
            feed = (yield (count))  # yields tock then waits for next send
            count += 1
            states.append(State(tyme=tymth(), context="recur", feed=feed, count = count))
            if count > 3:
                break  # normal exit

    except GeneratorExit:  # cease context, forced exit due to generator.close
        count += 1
        states.append(State(tyme=tymth(), context='cease', feed=feed, count=count))

    except Exception:  # abort context, forced exit due to uncaught exception
        count += 1
        states.append(State(tyme=tymth(), context='abort', feed=feed, count=count))
        raise

    finally:  # exit context,  unforced exit due to normal exit of try
        count += 1
        states.append(State(tyme=tymth(), context='exit', feed=feed, count=count))

    return (True)  # return value of yield from, or yield ex.value of StopIteration



# -*- encoding: utf-8 -*-
"""hio.help.mining module

Support for Mine dict subclass for shared in memory database
"""

from __future__ import annotations  # so type hints of classes get resolved later

import re
from collections.abc import Iterable, Mapping
from ..help import isNonStringIterable



# Regular expression to detect valid attribute names for Components that use Mine
# same as attribute name except '_' not allowed
NAMREX = r'^[a-zA-Z][a-zA-Z0-9]*$'
# Usage: if Renam.match(name): or if not Renam.match(name):
Renam = re.compile(NAMREX)  # compile is faster



class Mine(dict):
    """Mine subclass of dict with custom methods dunder methods and get that
    will only allow actual keys as str. Iterables passed in as key are converted
    to a "_' joined str. Uses "_" so can use dict constuctor if need be with str
    path. Assumes items in Iterable do not contain '_'.
    Supports attribute syntax to access items:
        mine.a = 5
        mine.a_b = 4

    Special staticmethods:
        tokeys(k) returns split of k at separator '_' as tuple.

    """
    def __init__(self, *pa, **kwa):
        """Convert keys that are tuples when positional argument is Iterable or
        Mapping to '.' joined strings

        dict __init__ signature options are:
            dict(**kwa)
            dict(mapping, **kwa)
            dict(iterable, **kwa)
        dict.update has same call signature
            d.update({"a": 5, "b": 2,}, c=3 , d=4)

        """
        self.update(*pa, **kwa)


    def __setitem__(self, k, v):
        return super(Mine, self).__setitem__(self.tokey(k), v)


    def __getitem__(self, k):
        return super(Mine, self).__getitem__(self.tokey(k))


    def __delitem__(self, k):
        return super(Mine, self).__delitem__(self.tokey(k))


    def __contains__(self, k):
        return super(Mine, self).__contains__(self.tokey(k))


    def __setattr__(self, k, v):
        try:
            getattr(dict, k)
        except AttributeError:
            try:
                return self.__setitem__(k, v)
            except Exception as ex:
                raise AttributeError(ex.args) from ex
        else:
            raise AttributeError(f"'{self.__class__.__name__}' attribute '{k}' "
                                 f"is read only")



    def __getattr__(self, k):
        if k not in self:
            raise AttributeError(f"{self.__class__.__name__} has no attribute '{k}'")
        return self.__getitem__(k)


    def get(self, k, default=None):
        if not self.__contains__(k):
            return default
        else:
            return self.__getitem__(k)



    def update(self, *pa, **kwa):
        """Convert keys that are tuples when positional argument is Iterable or
        Mapping to '.' joined strings

        dict __init__ signature options are:
            dict(**kwa)
            dict(mapping, **kwa)
            dict(iterable, **kwa)
        dict.update has same call signature
            d.update({"a": 5, "b": 2,}, c=3 , d=4)

        """
        if len(pa) > 1:
            raise TypeError(f"Expected 1 positional argument got {len(pa)}.")

        if pa:
            di = pa[0]
            if isinstance(di, Mapping):
                rd = {}
                for k, v in di.items():
                    rd[self.tokey(k)] = v
                super(Mine, self).update(rd, **kwa)

            elif isinstance(di, Iterable):
                ri = []
                for k, v in di:
                    ri.append((self.tokey(k), v))
                super(Mine, self).update(ri, **kwa)

            else:
                raise TypeError(f"Expected Mapping or Iterable got {type(di)}.")

        else:
            super(Mine, self).update(**kwa)


    @staticmethod
    def tokey(keys):
        """Joins tuple of strings keys to '_' joined string key. If already
        str then returns unchanged.

        Parameters:
            keys (Iterable[str] | str ): non-string Iteralble of path key
                    components to be '_' joined into key.
                    If keys is already str then returns unchanged

        Returns:
            key (str): '.' joined string
        """
        if isNonStringIterable(keys):
            try:
                for key in keys:
                    if key and not Renam.match(key):
                        raise KeyError(f"Invalid {key=}.")
                key = '_'.join(keys)
            except Exception as ex:
                raise KeyError(ex.args) from ex
        else:
            key = keys

        if not isinstance(key, str):
            raise KeyError(f"Expected str got {key}.")
        return key


    @staticmethod
    def tokeys(key):
        """Converts '_' joined string key to tuple of keys by splitting on '_'

        Parameters:
            key (str): '_' joined string to be split
        Returns:
            keys (tuple[str]): split of key on '_' into path key components
        """
        return tuple(key.split('_'))





# -*- encoding: utf-8 -*-
"""
hio.base Package
"""

from .tyming import Tymist, Tymee, Tymer
from .doing import Doist, doize, doify, Doer, DoDoer
from .filing import openFiler, Filer, FilerDoer
from .multidoing import Bosser, Crewer, TagDex



# -*- encoding: utf-8 -*-
"""
hio.core.memoing Module

Mixin Base Classes that add support for  memograms over datagram based transports.
In this context, a memogram is a larger memogram that is partitioned into
smaller parts as transport specific datagrams. This allows larger messages to
be transported over datagram transports than the underlying transport can support.

Provide mixin classes.
Takes of advantage of multiple inheritance to enable mixtures of behaviors
with minimal code duplication (more DRY).

New style python classes use the C3 linearization algorithm. Multiple inheritance
forms a directed acyclic graph called a diamond graph. This graph is linarized
into the method resolution order.
Use class.mro() or class.__mro__

(see https://www.geeksforgeeks.org/method-resolution-order-in-python-inheritance/)
Basically:
* children always precede their parents
* immediate parent classes of a child are visited in the order listed in the
child class statement.
* a super class is visited only after all sub classes have been visited
* linearized graph is monotonic (a class is only visted once)

Datagram IP Fragmentation
https://packetpushers.net/blog/ip-fragmentation-in-detail/
Only the first fragment contains the header that is needed to defragment at
the other end.

https://stackoverflow.com/questions/24045424/when-acting-on-a-udp-socket-what-can-cause-sendto-to-send-fewer-bytes-than-re
Answer is that UDP sendto always sends all the bytes but does not make a distinction
between nonblocking and blocking calls to sendto.
Other documentation seems to indicate that sendto when non-blocking may not
send all the bytes.
https://learn.microsoft.com/en-us/dotnet/api/system.net.sockets.socket.sendto?view=net-9.0

http://esp32.io/viewtopic.php?t=2378
For UDP, sendto() will try to allocate a packet buffer to hold the full
UDP packet of the size requested (up to 64KB) and will return with errno=ENOMEM
if this fails. Otherwise, provided FD_ISSET(sock, &write_set) is set then the
sendto() will succeed. But you should check for this failure.
Generally if the ESP32 is not low on free memory then you won't see ENOMEM here.

Every call to send must send less that sf::UdpSocket::MaxDatagramSize bytes --
which is a little less than 2^16 (65536) bytes.

https://people.computing.clemson.edu/~westall/853/notes/udpsend.pdf
However, the existence of the UDP and IPmid = uuid.uuid4().bytes  # 16 byte random uuid

headers should also limit.

Maximum IPV4 header size is 60 bytes and UDP header is 8 bytes so maximum UDP
datagram payload size is  65535 - 60 - 8 = 65467

With ipv6 the IP headers are done differently so maximum payload size is
65,535 - 8 = 65527

https://en.wikipedia.org/wiki/Maximum_transmission_unit
https://stackoverflow.com/questions/1098897/what-is-the-largest-safe-udp-packet-size-on-the-internet
The maximum safe UDP payload is 508 bytes. This is a packet size of 576
(the "minimum maximum reassembly buffer size"), minus the maximum 60-byte
IP header and the 8-byte UDP header. As others have mentioned, additional
protocol headers could be added in some circumstances. A more conservamid = uuid.uuid4().bytes  # 16 byte random uuid
tive
value of around 300-400 bytes may be preferred instead.

Usually ip headers are only 20 bytes so with UDP header (src, dst) of 8 bytes
the maximum allowed payload is 576 - 28 = 548

Any UDP payload this size or smaller is guaranteed to be deliverable over IP
(though not guaranteed to be delivered). Anything larger is allowed to be
outright dropped by any router for any reason.

Except on an IPv6-only route,
where the maximum payload is 1,212 bytes.

Unix Domain Socket:
https://unix.stackexchange.com/questions/424380/what-values-may-linux-use-for-the-default-unix-socket-buffer-size

Path cannot exceed 108 bytes
The  SO_SNDBUF  socket  option does have an effect for UNIX domain sockets,
but the SO_RCVBUF option does not.
For datagram sockets, the SO_SNDBUF value imposes an upper limit on the size
of outgoing datagrams.
This limit is calculated as the doubled (see socket(7)) option value less 32 bytes used for overhead.

Doubling accounts for the overhead.
SO_SNDBUF
              Sets or gets the maximum socket send buffer in bytes.  The
              kernel doubles this value (to allow space for bookkeeping
              overhead) when it is set using setsockopt(2), and this
              doubled value is returned by getsockopt(2).  The default
              value is set by the /proc/sys/net/core/wmem_default file
              and the maximum allowed value is set by the
              /proc/sys/net/core/wmem_max file.  The minimum (doubled)
              value for this option is 2048.

In python the getsockopt apparently does not return the doubled value but
the undoubled value.

https://unix.stackexchange.com/questions/38043/size-of-data-that-can-be-written-to-read-from-sockets
https://stackoverflow.com/questions/21856517/whats-the-practical-limitmid = uuid.uuid4().bytes  # 16 byte random uuid
-on-the-size-of-single-packet-transmitted-over-domain

Memo
Memo partitioning into grams.

Default gram is of form  gram = head + sep + body.
Some subclasses might have gram of form gram = head + sep + body + tail.
In that case encoding of body + tail must provide a way to separate body from tail.
Typically tail would be a signature on the fore gram = head + sep + body


Separator sep is b'|' must not be a base64 character.

The head consists of three fields in base64
mid = memo IDmid = uuid.uuid4().bytes  # 16 byte random uuid

gn = gram number of gram in memo zero based
gc = gram count of total grams in memo may be zero when body is empty

Sep = b'|'
assert not helping.Reb64.match(Sep)

body = b"demogram"
gn = 0
gc = 12

gn = helping.intToB64b(gn, l=4)
gc = helping.intToB64b(gc, l=4)

GramLeader = struct.Struct('!16s4s4s')
GramLeader.size  == 24
head = GramLeader.pack(mid, gn, gc)
gram = Sep.join(head, body)


head, sep, body = gram.partition(Sep)
assert helping.Reb64.match(head)
mid, gn, gc = GramLeader.unpack(head)
gn = helping.b64ToInt(gn)
gc = helping.b64ToInt(gc)

# test GramHead
code = MtrDex.GramHead
assert code == '0P'
codeb = code.encode()mid = uuid.uuid4().bytes  # 16 byte random uuid



mid = uuid.uuid4().bytes  # 16 byte random uuid


mid = 1
midb = mid.to_bytes(16)
assert midb == b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01'
gn = 1
gnb = gn.to_bytes(3)
assert gnb == b'\x00\x00\x01'
gc = 2
gcb = gc.to_bytes(3)
assert gcb == b'\x00\x00\x02'
raw = midb + gnb + gcb
assert raw == (b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01'
               b'\x00\x00\x01\x00\x00\x02')mid = uuid.uuid4().bytes  # 16 byte random uuid


assert mid == int.from_bytes(raw[:16])
assert gn == int.from_bytes(raw[16:19])
assert gc == int.from_bytes(raw[19:22])

midb64b = encodeB64(bytes([0] * 2) + midb)[2:] # prepad ans strip
gnb64b = encodeB64(gnb)
gcb64b = encodeB64(gcb)

qb64b = codeb + midb64b + gnb64b + gcb64b
assert qb64b == b'0PAAAAAAAAAAAAAAAAAAAAABAAABAAAC'
qb64 = qb64b.decode()
qb2 = decodeB64(qb64b)

assert mid == int.from_bytes(decodeB64(b'AA' + qb64b[2:24]))
assert gn == int.from_bytes(decodeB64(qb64b[24:28]))
assert gc == int.from_bytes(decodeB64(qb64b[28:32]))


codeb64b = b'__"

# 16 byte random uuid
midb64b = encodeB64(bytes([0] * 2) + uuid.uuid4().bytes)[2:] # prepad then strip
gn = 1
gnb64b = helping.intToB64b(gn, l=3)
gc = 2
gcb64b = helping.intToB64b(gc, l=3)

headb64b = codeb64b + midb64b + gnb64b + gcb64b

fmt = '!2s22s4s4s'
GramLeader = struct.Struct(fmt)
GramLeader.size  == 32
head = GramLeader.pack(codeb64b, midb64b, gnb64b, gcb64b)

assert helping.Reb64.match(head)
codeb64b, midb64b, gnb64b, gcb64b = GramLeader.unpack(head)
codeb64b, midb64b, gnb64b, gcb64b = GramLeader unpack_from(fmt, gram)


mid = helping.b64ToInt(b'AA' + midb64b))
gn = helping.b64ToInt(gnb64b))
gc = helping.b64ToInt(gcb64b))

Standard header in all grams consists of
code + mid + gn

The first gram with gn == 0 also has an additional field called the neck
that is the total gram count gc

Standard
HeadSize = 28  # base64 chars or 21 base2 bytes
NeckSize = 4 # base  64 chars or 21 base2 bytes

codeb64b = b'__"
midb64b = encodeB64(bytes([0] * 2) + uuid.uuid4().bytes)[2:] # prepad, convert,  and strip
gn = 1
gnb64b = helping.intToB64b(gn, l=4)
gc = 2
gcb64b = helping.intToB64b(gc, l=4)

headb64b = codeb64b + midb64b + gnb64b
neckb64b = gcb64b
assert helping.Reb64.match(headb64b)
assert helping.Reb64.match(neckb64b)

codeb64b = headb64b[:2]
mid = helping.b64ToInt(b'AA' + headb64b[2:24]))  # prepad and convert
gn = helping.b64ToInt(headb64b[24:28]))
gc = helping.b64ToInt(neckb64b[28:32]))


GramCode Typing/Versions uses a different base64 two char code.
Codes always start with `_` and second char in code starts at `_`
and walks backwards through B64 code table for each gram type-version (tv) code.
Each of total head length including code and total neck length must be a
multiple of 4 characters.  so eacch converts to/from B2 without padding.
Head type-version code governs the neck as well. So to change length in order to change
field lengths or add or remove fields need a new gram (tv) code.

To reiterate, if ever need to change anything ned a new gram tv code. There are
no versions per type.  This is deemed sufficient because anticipate a very
limited set of possible fields ever needed for memogram transport types.

tv 0 is '__'
tv 1 is '_-'
tv 3 is '_9'

Because the gram head and neck are valid mid-pad B64 primitives then can be losslessly
transformed to/from CESR primitives merely by translated the gram tv code to
an  equivalent entry in the  two char cesr primitive code table. When doing so
The neck is always attached and then stripped when not needed.

The CESR codes albeit different are one-to-one. This enables representing headers
as CESR primitives for management but not over the wire. I.e. can transform
Memogram Grams to CESR primitives to tunnel in some CESR stream.

Normally CESR streams are tunneled inside Memoer memograms sent over-the-wire.
In this case, the memogram is the wrapper and each gram uses the gram tv code
defined herein not the equivalent CESR code.
Using  so far reserved and unused '_' code which is reserved as the selector
for theCESR op code table for memogram grams makes it more apparent that a gram
belongs to a memogram and not a CESR stream.
Eventually when CESR op codes eventually become defined, it's not likely to be
confusing since the context of CESR op codes is different that transport
wrappers.

Morover when a MemoGram payload is a tunneled CESR Stream, the memogram parser
will parse and strip the MemoGram gram wrappers to construct the tunneled,
stream so the wrapper is transarent to the CESR stream and the CESR stream
payload is opaque to the memogram wrapper, i.e. no ambiguity.



HeadCode hc = b'__'
HeadCodeSize hcs = 2
MemoIDSize mis = 22
GramNumSize gns = 4
GramCntSize gcs = 4
GramHeadSize phs  = 28
GramHeadNeckSize phns = 28 + 4 = 32


MaxGramBodySize = ()2** 16 -1) - HeadSize
GramBodySize pbs = something <= MaxBodySize
GramSize = GramHeadSize + GramBodySize
GramSize <= 2**16 -1

MaxMemoSize = (2**16-1)**2
MaxGramCount for UDP GramSize of 508 is ceil(MaxMemoSize/508) = 8454403  which is < 2**24-1
not gramsize includes head
MaxGramCount for UXD GramSize of 2**16 -1  is ceil(MaxMemoSize/(2**16-1))= 2**16-1
MinGramSize = MaxMemoSize // (2**24-1) = 255  = 2**8-1
MaxGramSize = (2**16-1)
For any given GramSize  there is a MaxGramCount of 2**24-1 so if fix the GramSize
this limits the effective
mms = min((2**16-1)**2), (GramSize - GramHeadSize) * (2**24-1))
mms = min((2**16-1)**2), (GramBodySize) * (2**24-1))
Note there is an extra 4 characters in first gram for the neck

so ps, pbs, and mms are variables specific to transport
PHS is fixed for the transport type reliable unreliable with header fields as defined
The desired ps for a given transport instance may be smaller than the allowed
maximum to accomodate buffers etc.
Given the fixed gram head size PHS once can calculate the maximum memo size
that includes the header overhead given the contraint of no more than 2**24-1
grams in a given memo.

So for a given transport:
ps == min(ps, 2**16-1)
pbs = ps - PHS

mms = min((2**16-1)**2), pbs * (2**24-1))
So compare memo size to mms and if greater raise error and drop memo
otherwise partition (gram) memo into grams with headers and transmit


Note max memo size is (2**16-1)**2 = 4294836225.
But accounting for the gram header overhead the largest payload max memo size is
max gram size = (2**16-1) == 65535
max gram count = (2**24-1) == 16777215
standard header size hs = 28
neck size ns = 4
bs = maxgramsize - hs = (65535 - hs) = 65507  # max standard gram body size
max theoretical memogram payload size mms is
mms = (bs * maxgramcount) - ns  = 65507 * 16777215 - 4 = 1099025023001
but this is too big so we limit it to 4294836225 but we can have a payload
of 4294836225 when using the maxgramsize for each gram.

CESR groups code count in quadlets (4 chars).
The big group count codes have five digits of 6 bits each or 30 bits to count the
size of the following framed chars.
The max count value is therefore (2**30-1). Cesr group codes count the number
of quadlets (4 chars) in the framed group. The maximum number of bytes in a
CESR big group frame body is therefore:
(2**30-1)*4 = 4294967292.
The group code is not gram of the counted quadlets in the frame body.
The big group codes are 8 chars long. Thus with its group code the total big
frame size is:
(2**30-1)*4 + 8 = 4294967300

The largest value for a 32 bit number is
(2 ** 32 - 1) = 4294967295 which is 5 less than the largest group frame size with code.
4294967300 - 4294967295 = 5
Largest total frame size with group code = 2**32+4 which exceeds the largest
value of a 32 bit number.

Restricting the largest memogram payload to be 2**32-1 should be sufficient.
It is anticipated that other limits such as memory or processing power will
constrain memogram sizes to be much smaller than 2**32-1 much less 2**32+4.
So MaximumMemoSize is 2**32-1 = 4294967295
MaxGramSize with overhead is 2**16-1 = 65535 which is a limit on some datagram
transports for the maximum sized datagram. This is true of UDP although the
UDP MTU size is much smaller than this. It may not longer be true on some
OSes that the maximum datagram size for UXD transports is 65535. On some
OSes it may be bigger. But this seems like a good practical limit nonetheless
as OS buffer sizes are usually in the order of magnitude of 65535.


The maximum group count is (2**30-1) == 1073741823




ToDo:
Make gram body size calculation a function of the current .curt for header
encoding as b2 headers are smaller so the body payload is bigger.

Add code '_-' for signed grams   vs verification id size ss signatures size
expanes Sizes for vs and ss
Signed first gram  code + mid + vid + gn + gc + body + sig
Signed other grams  code + mid + vid + gn + body + sig
Add sign and verify methods



"""
import socket
import errno
import math
import uuid
#import struct

from collections import deque, namedtuple
from contextlib import contextmanager
from base64 import urlsafe_b64encode as encodeB64
from base64 import urlsafe_b64decode as decodeB64
from dataclasses import dataclass, astuple, asdict

from ... import hioing, help
from ...base import tyming, doing
from ...help import helping

logger = help.ogler.getLogger()

# namedtuple of ints (major: int, minor: int)
Versionage = namedtuple("Versionage", "major minor")


# namedtuple for gram header part size entries in Memoer code tables
# cs is the code part size int number of chars in code part
# ms is the mid part size int number of chars in the mid part (memoID)
# vs is the vid  part size int number of chars in the vid part (verification ID)
# ns is the neck part size int number of chars for the gram number in all grams
#       and the additional neck that also appears in first gram for gram count
# ss is the signature part size int number of chars in the signature part.
#       the signature part is discontinuously attached after the body part but
#       its size is included in over head computation for the body size
# hs is the head size int number of chars for other grams no-neck.
#       hs = cs + ms + vs + ns + ss. The size for the first gram with neck is
#       hs + ns


Sizage = namedtuple("Sizage", "cs ms vs ss ns hs")

@dataclass(frozen=True)
class GramCodex:
    """
    GramCodex is codex of all Gram Codes.
    Only provide defined codes.
    Undefined are left out so that inclusion(exclusion) via 'in' operator works.
    """
    Basic:     str = '__'  # Basic gram code for basic overhead parts
    Signed:    str = '_-'  # Basic gram code for signed overhead parts

    def __iter__(self):
        return iter(astuple(self))


GramDex = GramCodex()  # Make instance

@dataclass(frozen=True)
class SignGramCodex:
    """
    SignGramCodex is codex of all Signed Gram Codes.
    Only provide defined codes.
    Undefined are left out so that inclusion(exclusion) via 'in' operator works.
    """
    Signed:    str = '_-'  # Basic gram code for signed overhead parts

    def __iter__(self):
        return iter(astuple(self))


SGDex = SignGramCodex()  # Make instance


class Memoer(hioing.Mixin):
    """
    Memoer mixin base class to adds memogram support to a transport class.
    Memoer supports asynchronous memograms. Provides common methods for subclasses.

    A memogram is a higher level construct that sits on top of a datagram.
    A memogram supports the segmentation and desegmentation of memos to
    respect the underlying datagram size, buffer, and fragmentation behavior.

    Layering a reliable transaction protocol on top of a memogram enables reliable
    asynchronous messaging over unreliable datagram transport protocols.

    When the datagram protocol is already reliable, then a memogram enables
    larger memos (messages) than that natively supported by the datagram.

    Usage:
        Do not instantiate directly but use as a mixin with a transport class
        in order to create a new subclass that adds memogram support to the
        transport class. For example MemoerUdp or MemoerUxd

    Each direction of dataflow uses a tiered set of buffers that respect
    the constraints of non-blocking asynchronous IO with datagram transports.

    On the receive side each complete datagram (gram) is put in a gram receive
    deque as a segment of a memo. These deques are indexed by the sender's
    source addr. The grams in the gram recieve deque are then desegmented into a memo
    and placed in the memo deque for consumption by the application or some other
    higher level protocol.

    On the transmit side memos are placed in a deque (double ended queue). Each
    memo is then segmented into grams (datagrams) that respect the size constraints
    of the underlying datagram transport. These grams are placed in the outgoing
    gram deque. Each entry in this deque is a duple of form:
    (gram: bytes, dst: str). Each  duple is pulled off the deque and its
    gram is put in bytearray for transport.

    When using non-blocking IO, asynchronous datagram transport
    protocols may have hidden buffering constraints that result in fragmentation
    of the sent datagram which means the whole datagram is not sent at once via
    a non-blocking send call. This means that the remainer of the datagram must
    be sent later and may take multiple send calls to complete. The datagram
    protocol is responsible for managing the buffering and fragmentation but
    depends on the sender repeated attempts to send the reaminder of the
    full datagram. This is ensured with the final tier with a raw transmit
    buffer that waits until it is empty before attempting to send another
    gram. Because sending to different destinations may fail for different reasons
    such as bad addresses or bad routing each destination gets its own buffer
    so that a bad destination does not block other destinations.

    Memo segmentation/desegmentation information is embedded in the grams.

    Inherited Class Attributes:
        MaxGramSize (int): absolute max gram size on tx with overhead

    Class Attributes:
        Version (Versionage): default version consisting of namedtuple of form
            (major: int, minor: int)
        Codex (GramDex): dataclass ref to gram codex
        Codes (dict): maps codex names to codex values
        Names (dict): maps codex values to codex names
        Sodex (SGDex): dataclass ref to signed gram codex
        Sizes (dict): gram head part sizes Sizage instances keyed by gram codes
        MaxMemoSize (int): absolute max memo size
        MaxGramCount (int): absolute max gram count


    Inherited Attributes:
        name (str):  unique name for Memoer transport. Used to manage.
        opened (bool):  True means transport open for use
                        False otherwise
        bc (int | None): count of transport buffers of MaxGramSize

    Attributes:
        version (Versionage): version for this memoir instance consisting of
                namedtuple of form (major: int, minor: int)
        rxgs (dict): holding rx (receive) (data) gram deques of grams.
            Each item in dict has key=src and val=deque of grames received
        rxgs (dict): keyed by mid (memoID) with value of dict where each deque
            holds incomplete memo grams for that mid.
            The mid appears in every gram from the same memo.
            The value dict is keyed by the gram number with value
            that is the gram bytes.
        sources (dict): keyed by mid (memoID) that holds the src for the memo.
            This enables reattaching src to fused memo in rxms deque tuple.
        counts (dict): keyed by mid (memoID) that holds the gram count from
            the first gram for the memo. This enables lookup of the gram count when
            fusing its grams.
        vids (dict[mid: (vid | None)]): keyed by mid that holds the verification ID str for
                the memo indexed by its mid (memoID). This enables reattaching
                the vid to memo when placing fused memo in rxms deque.
                Vid is only present when signed header otherwise vid is None
        rxms (deque): holding rx (receive) memo tuples desegmented from rxgs grams
                each entry in deque is tuple of form:
                (memo: str, src: str, vid: str) where:
                memo is fused memo, src is source addr, vid is verification ID
        txms (deque): holding tx (transmit) memo tuples to be segmented into
                txgs grams where each entry in deque is tuple of form
                (memo: str, dst: str, vid: str | None)
                memo is memo to be partitioned into gram
                dst is dst addr for grams
                vid is verification id when gram is to be signed or None otherwise
        txgs (deque): grams to transmit, each entry is duple of form:
                (gram: bytes, dst: str).
        txbs (tuple): current transmisstion duple of form:
            (gram: bytearray, dst: str). gram bytearray may hold untransmitted
            portion when Encodesdatagram is not able to be sent all at once so can
            keep trying. Nothing to send indicated by (bytearray(), None)
            for (gram, dst)
        echos (deque): holding echo receive duples for testing. Each duple of
                       form: (gram: bytes, dst: str).


    Hidden:
        _code (bytes | None): see size property
        _curt (bool): see curt property
        _size (int): see size property
        _verific (bool): see verific property



    Properties:
        code (bytes | None): gram code for gram header when rending for tx
        curt (bool): True means when rending for tx encode header in base2
                     False means when rending for tx encode header in base64
        size (int): gram size when rending for tx.
            first gram size = over head size + neck size + body size.
            other gram size = over head size + body size.
            Min gram body size is one.
            Gram size also limited by MaxGramSize and MaxGramCount relative to
            MaxMemoSize.
        verific (bool): True means any rx grams must be signed.
                        False otherwise


    """
    Version = Versionage(major=0, minor=0)  # default version
    Codex = GramDex
    Codes = asdict(Codex)  # map code name to code
    Names = {val : key for key, val in Codes.items()} # invert map code to code name
    Sodex = SGDex  # signed gram codex
    # dict of gram header part sizes keyed by gram codes: cs ms vs ss ns hs
    Sizes = {
                '__': Sizage(cs=2, ms=22, vs=0, ss=0, ns=4, hs=28),
                '_-': Sizage(cs=2, ms=22, vs=44, ss=88, ns=4, hs=160),
             }
    #Bodes = ({helping.codeB64ToB2(c): c for n, c in Codes.items()})
    MaxMemoSize = 4294967295 # (2**32-1) absolute max memo payload size
    MaxGramCount = 16777215 # (2**24-1) absolute max gram count
    MaxGramSize = 65535  # (2**16-1)  Overridden in subclass


    def __init__(self, *,
                 name=None,
                 bc=None,
                 version=None,
                 rxgs=None,
                 sources=None,
                 counts=None,
                 vids=None,
                 rxms=None,
                 txms=None,
                 txgs=None,
                 txbs=None,
                 code=GramDex.Basic,
                 curt=False,
                 size=None,
                 verific=False,
                 **kwa
                ):
        """Setup instance

        Inherited Parameters:
            name (str): unique name for Memoer transport. Used to manage.
            opened (bool): True means opened for transport
                           False otherwise
            bc (int | None): count of transport buffers of MaxGramSize

        Parameters:
            version (Versionage): version for this memoir instance consisting of
                namedtuple of form (major: int, minor: int)
            rxgs (dict): keyed by mid (memoID) with value of dict where each
                value dict holds grams from memo keyed by gram number.
                Grams have been stripped of their headers.
                The mid appears in every gram from the same memo.
                The value dict is keyed by the gram number gn, with value bytes.
            sources (dict): keyed by mid that holds the src for the memo indexed
                by its mid (memoID). This enables reattaching src to memo when
                placing fused memo in rxms deque.
            counts (dict): keyed by mid that holds the gram count for the
                memo indexed by its mid (memoID). This enables lookup of the
                gram count for a given memo to know when it has received all its
                constituent grams in order to fuse back into the memo.
            vids (dict[mid: (vid | None)]): keyed by mid that holds the verification ID str for
                the memo indexed by its mid (memoID). This enables reattaching
                the vid to memo when placing fused memo in rxms deque.
                Vid is only present when signed header otherwise vid is None
            rxms (deque): holding rx (receive) memo tuples desegmented from rxgs grams
                each entry in deque is tuple of form:
                (memo: str, src: str, vid: str) where:
                memo is fused memo, src is source addr, vid is verification ID
            txms (deque): holding tx (transmit) memo tuples to be segmented into
                txgs grams where each entry in deque is tuple of form
                (memo: str, dst: str, vid: str | None)
                memo is memo to be partitioned into gram
                dst is dst addr for grams
                vid is verification id when gram is to be signed or None otherwise
            txgs (deque): grams to transmit, each entry is duple of form:
                (gram: bytes, dst: str). Grams include gram headers.
            txbs (tuple): current transmisstion duple of form:
                (gram: bytearray, dst: str). gram bytearray may hold untransmitted
                portion when datagram is not able to be sent all at once so can
                keep trying. Nothing to send indicated by (bytearray(), None)
                for (gram, dst)
            code (bytes): gram code for gram header
            curt (bool): True means when rending for tx encode header in base2
                         False means when rending for tx encode header in base64
            size (int): gram size when rending for tx.
                first gram size = head size + neck size + body size.
                other gram size = head size + body size.
                Min gram body size is one.
                Gram size also limited by MaxGramSize and MaxGramCount relative to
                MaxMemoSize.
            verific (bool): True means any rx grams must be signed.
                            False otherwise
        """

        # initialize attributes
        self.version = version if version is not None else self.Version
        self.rxgs = rxgs if rxgs is not None else dict()
        self.sources = sources if sources is not None else dict()
        self.counts = counts if counts is not None else dict()
        self.vids = vids if vids is not None else dict()
        self.rxms = rxms if rxms is not None else deque()

        self.txms = txms if txms is not None else deque()
        self.txgs = txgs if txgs is not None else deque()
        self.txbs = txbs if txbs is not None else (bytearray(), None)

        self.echos = deque()  # only used in testing as echoed tx

        self.code = code
        self.curt = curt
        self.size = size  # property sets size given .code and constraints
        self._verific = True if verific else False

        super(Memoer, self).__init__(name=name, bc=bc, **kwa)

        if not hasattr(self, "name"):  # stub so mixin works in isolation.
            self.name = name if name is not None else "main"  # mixed with subclass should provide this.

        if not hasattr(self, "opened"):  # stub so mixin works in isolation.
            self.opened = False  # mixed with subclass should provide this.

        if not hasattr(self, "bc"):  # stub so mixin works in isolation.
            self.bc = bc if bc is not None else 4  # mixed with subclass should provide this.


    @property
    def code(self):
        """Property getter for ._code

        Returns:
            code (str): two char base64 gram code
        """
        return self._code


    @code.setter
    def code(self, code):
        """Property setter for ._code

        Parameters:
            code (str): two char base64 gram code
        """
        if code not in self.Codex:
            raise hioing.MemoerError(f"Invalid {code=}.")

        self._code = code
        if hasattr(self, "_size"):
            self.size = self._size  # refresh size given new code


    @property
    def curt(self):
        """Property getter for ._curt

        Returns:
            curt (bool): True means when rending for tx encode header in base2
                         False means when rending for tx encode header in base64
        """
        return self._curt


    @curt.setter
    def curt(self, curt):
        """Property setter for ._curt

        Parameters:
            curt (bool): True means when rending for tx encode header in base2
                         False means when rending for tx encode header in base64
        """
        self._curt = curt
        if hasattr(self, "_size"):
            self.size = self._size  # refresh size given new curt


    @property
    def size(self):
        """Property getter for ._size

        Returns:
            size (int): gram size when rending for tx.
                        First gram size = head size + neck size + body size.
                        Other gram size = head size + body size.
                        Min gram body size is one.
                        Gram size also limited by MaxGramSize and MaxGramCount
                        relative to MaxMemoSize.
        """
        return self._size


    @size.setter
    def size(self, size):
        """Property setter for ._size

        Parameters:
            size (int | None): gram size for rending memo

        """
        _, _, _, _, ns, hs = self.Sizes[self.code]  # cs ms vs ss ns hs
        size = size if size is not None else self.MaxGramSize
        if self.curt:  # minimum header smaller when in base2 curt
            hs = 3 * hs // 4
            ns = 3 * ns // 4

        # mininum size must be big enough for first gram header and 1 body byte
        self._size = max(min(size, self.MaxGramSize), hs + ns + 1)


    @property
    def verific(self):
        """Property getter for ._verific

        Returns:
            verific (bool): True means any rx grams must be signed.
                            False otherwise
        """
        return self._verific


    def open(self):
        """Opens transport in nonblocking mode

        This is a stub. Override in transport specific subclass
        """
        self.opened = True
        return True


    def reopen(self, **kwa):
        """Idempotently open transport

        This is a stub. Override in transport specific subclass
        """
        self.close()
        return self.open()


    def close(self):
        """Closes  transport

        This is a stub. Override in transport subclass
        """
        self.opened = False


    def wiff(self, gram):
        """Determines encoding of gram bytes header when parsing grams.
        The encoding maybe either base2 or base64.


        Returns:
            curt (bool):    True means base2 encoding
                            False means base64 encoding
                            Otherwise raises hioing.MemoerError

        All gram head codes start with '_' in base64 text or in base2 binary.
        Given only allowed chars are from the set of base64 then can determine
        if header is in base64 or base2.

        First sextet:  (0b is binary, 0o is octal, 0x is hexadecimal)

        0o27 = 0b010111 means first sextet of '_' in base64
        0o77 = 0b111111 means first sextet of '_' in base2

        Assuming that only '_' is allowed as the first char of a valid gram head,
        in either Base64 or Base2 encoding, a parser can unambiguously determine
        if the gram header encoding is binary Base2 or text Base64 because
        0o27 != 0o77.

        Furthermore, it just so happens that 0o27 is a unique first sextet
        among Base64 ascii chars. So an invalid gram header when in Base64 encoding
        is detectable. However, there is one collision (see below) with invalid
        gram headers in Base2 encoding. So an erroneously constructed gram might
        not be detected merely by looking at the first sextet. Therefore
        unreliable transports must provide some additional mechanism such as a
        hash or signature on the gram.

        All Base2 codes for Base64 chars are unique since the B2 codes are just
        the count from 0 to 63. There is one collision, however, between Base2 and
        Base 64 for invalid gram headers. This is because 0o27 is the
        base2 code for ascii 'V'. This means that Base2 encodings
        that begin 0o27 witch is the Base2 encoding of the ascii 'V, would be
        incorrectly detected as text not binary.
        """

        sextet = gram[0] >> 2
        if sextet == 0o27:
            return False  # base64 text encoding
        if sextet == 0o77:
            return True  # base2 binary encoding

        raise hioing.MemoerError(f"Unexpected {sextet=} at gram head start.")


    def verify(self, sig, ser, vid):
        """Verify signature sig on signed part of gram, ser, using key from vid.
        Must be overriden in subclass to perform real signature verification.
        This is a stub.

        Returns:
            result (bool): True if signature verifies
                           False otherwise

        Parameters:
            sig (bytes | str): qualified base64 qb64b
            ser (bytes): signed portion of gram in delivered format
            vid (bytes | str): qualified base64 qb64b of verifier ID
        """
        if hasattr(sig, 'encode'):
            sig = sig.encode()

        if hasattr(vid, 'encode'):
            vid = vid.encode()

        return True


    def pick(self, gram):
        """Strips header from gram bytearray leaving only gram body in gram and
        returns (mid, gn, gc). Raises MemoerError if unrecognized header

        Returns:
            result (tuple): tuple of form:
                (mid: str, vid: str, gn: int, gc: int | None) where:
                mid is fully qualified memoID,
                vid is verID,
                gn is gram number,
                gc is gram count.
                When first gram (zeroth) returns (mid, vid, 0, gc).
                When other gram returns (mid, vid, gn, None)
                When code has empty vid then vid is None
                Otherwise raises MemoerError error.

        When valid recognized header, strips header bytes from front of gram
        leaving the gram body part bytearray.

        Parameters:
            gram (bytearray): memo gram from which to parse and strip its header.


        """
        curt = self.wiff(gram)  # rx gram encoding True=B2 or False=B64
        if curt:  # base2 binary encoding
            if len(gram) < 2:  # assumes len(code) must be 2
                raise hioing.MemoerError(f"Gram length={len(gram)} to short to "
                                         f"hold code.")
            code = helping.codeB2ToB64(gram, 2)  # assumes len(code) must be 2
            if self.verific and code not in self.Sodex:  # must be signed
                raise hioing.MemoerError(f"Unsigned gram {code =} when signed "
                                         f"required.")

            cs, ms, vs, ss, ns, hs = self.Sizes[code]  # cs ms vs ss ns hs
            ps = (3 - ((ms) % 3)) % 3  # net pad size for mid
            cms = 3 * (cs + ms) // 4  # cs + ms are aligned on 24 bit boundary
            hs = 3 * hs // 4  # encoding b2 means head part sizes smaller by 3/4
            ns = 3 * ns // 4  # encoding b2 means head part sizes smaller by 3/4
            vs = 3 * vs // 4  # encoding b2 means head part sizes smaller by 3/4
            ss = 3 * ss // 4  # encoding b2 means head part sizes smaller by 3/4

            if len(gram) < (hs + 1):  # not big enough for non-first gram
                raise hioing.MemoerError(f"Not enough rx bytes for b2 gram"
                                         f" < {hs + 1}.")

            #mid = encodeB64(bytes([0] * ps) + gram[cs:cms])[ps:].decode()  # prepad, convert, strip
            mid = encodeB64(gram[:cms]).decode()  # fully qualified with prefix code
            vid = encodeB64(gram[cms:cms+vs]).decode()  # must be on 24 bit boundary
            gn = int.from_bytes(gram[cms+vs:cms+vs+ns])
            if gn == 0:  # first (zeroth) gram so get neck
                if len(gram) < hs + ns + 1:
                    raise hioing.MemoerError(f"Not enough rx bytes for b2 "
                                             f"gram < {hs + ns + 1}.")
                neck = gram[cms+vs+ns:cms+vs+2*ns]  # slice takes a copy
                gc = int.from_bytes(neck)  # convert to int
                sig = encodeB64(gram[-ss if ss else len(gram):])   # last ss bytes are signature
                del gram[-ss if ss else len(gram):]  # strip sig
                signed = bytes(gram[:])  # copy signed portion of gram
                del gram[:hs-ss+ns]  # strip of fore head leaving body in gram
            else:  # non-first gram no neck
                gc = None
                sig = encodeB64(gram[-ss if ss else len(gram):])
                del gram[-ss if ss else len(gram):]  # strip sig
                signed = bytes(gram[:])  # copy signed portion of gram
                del gram[:hs-ss]  # strip of fore head leaving body in gram

        else:  # base64 text encoding
            if len(gram) < 2:  # assumes len(code) must be 2
                raise hioing.MemoerError(f"Gram length={len(gram)} to short to "
                                         f"hold code.")
            code = gram[:2].decode()  # assumes len(code) must be 2
            if self.verific and code not in self.Sodex:  # must be signed
                raise hioing.MemoerError(f"Unsigned gram {code =} when signed "
                                         f"required.")
            cs, ms, vs, ss, ns, hs = self.Sizes[code]  # cs ms vs ss ns hs

            if len(gram) < (hs + 1):  # not big enough for non-first gram
                raise hioing.MemoerError(f"Not enough rx bytes for b64 gram"
                                         f" < {hs + 1}.")

            mid = bytes(gram[:cs+ms]).decode()  # fully qualified with prefix code
            vid = bytes(gram[cs+ms:cs+ms+vs]).decode() # must be on 24 bit boundary
            gn = helping.b64ToInt(gram[cs+ms+vs:cs+ms+vs+ns])
            if gn == 0:  # first (zeroth) gram so get neck
                if len(gram) < hs + ns + 1:
                    raise hioing.MemoerError(f"Not enough rx bytes for b64 "
                                             f"gram < {hs + ns + 1}.")
                neck = gram[cs+ms+vs+ns:cs+ms+vs+2*ns]  # slice takes a copy
                gc = helping.b64ToInt(neck)  # convert to int
                sig = bytes(gram[-ss if ss else len(gram):])   # last ss bytes are signature
                del gram[-ss if ss else len(gram):]  # strip sig
                signed = bytes(gram[:])  # copy signed portion of gram
                del gram[:hs-ss+ns]  # strip of fore head leaving body in gram
            else:  # non-first gram no neck
                gc = None
                sig = bytes(gram[-ss if ss else len(gram):])
                del gram[-ss if ss else len(gram):]  # strip sig
                signed = bytes(gram[:])  # copy signed portion of gram
                del gram[:hs-ss]  # strip of fore head leaving body in gram

        if sig:  # signature not empty
            if not self.verify(sig, signed, vid):
                raise hioing.MemoerError(f"Invalid signature on gram from "
                                         f"verifier {vid=}.")

        return (mid, vid if vid else None, gn, gc)


    def receive(self, *, echoic=False):
        """Attemps to send bytes in txbs to remote destination dst.
        Must be overridden in subclass.
        This is a stub to define mixin interface.

        Parameters:
            echoic (bool): True means use .echos in .receive debugging purposes
                            where echo is duple of form: (gram: bytes, src: str)
                           False means do not use .echos default is duple that
                            indicates nothing to receive of form (b'', None)

        Returns:
            duple (tuple): of form (data: bytes, src: str) where data is the
                bytes of received data and src is the source address.
                When no data the duple is (b'', None) unless echoic is True
                then pop off echo from .echos
        """

        if echoic:
            try:
                result = self.echos.popleft()
            except IndexError:
                result = (b'', None)
        else:
            result = (b'', None)

        return result


    def _serviceOneReceived(self, *, echoic=False):
        """Service one received duple (raw, src) raw packet data. Always returns
        complete datagram.

        Returns:
            result (bool):  True means data received from source over transport
                            False means no data so try again later

                        return enables greedy callers to keep calling until no
                        more data to receive from transport

        Parameters:
            echoic (bool): True means use .echos in .receive debugging purposes
                            where echo is duple of form: (gram: bytes, src: str)
                           False means do not use .echos default is duple that
                            indicates nothing to receive of form (b'', None)

        """
        try:
            gram, src = self.receive(echoic=echoic)  # if no data the duple is (b'', None)
        except socket.error as ex:  # OSError.errno always .args[0] for compat
            #if (ex.args[0] == (errno.ECONNREFUSED,
                               #errno.ECONNRESET,
                               #errno.ENETRESET,
                               #errno.ENETUNREACH,
                               #errno.EHOSTUNREACH,
                               #errno.ENETDOWN,
                               #errno.EHOSTDOWN,
                               #errno.ETIMEDOUT,
                               #errno.ETIME,
                               #errno.ENOBUFS,
                               #errno.ENOMEM)):
                #return False  # no received data
            #else:
            raise  # should not happen

        if not gram:  # no received data
            return False  # so try again later

        gram = bytearray(gram)# make copy bytearray so can strip off header

        try:
            mid, vid, gn, gc = self.pick(gram)  # parse and strip off head leaving body
        except hioing.MemoerError as ex: # invalid gram so drop
            logger.error("Unrecognized Memoer gram from %s.\n %s.", src, ex)
            return True  # did receive data so can try again now

        if mid not in self.rxgs:
            self.rxgs[mid] = dict()

        # save stripped gram to be fused later
        if gn not in self.rxgs[mid]:  # make idempotent first only no replay
            self.rxgs[mid][gn] = gram  # index body by its gram number

        if gc is not None:
            if mid not in self.counts:  # make idempotent first only no replay
                self.counts[mid] = gc  # save gram count for mid

        if mid not in self.vids:
            self.vids[mid] = vid

        # assumes unique mid across all possible sources. No replay by different
        # source only first source for a given mid is ever recognized
        if mid not in self.sources:  # make idempotent first only no replay
            self.sources[mid] = src  # save source for later

        return True  # received valid so can try again now


    def serviceReceivesOnce(self, *, echoic=False):
        """Service receives once (non-greedy) and queue up

        Parameters:
            echoic (bool): True means use .echos in .receive debugging purposes
                            where echo is duple of form: (gram: bytes, src: str)
                           False means do not use .echos default is duple that
                            indicates nothing to receive of form (b'', None)
        """
        if self.opened:
            self._serviceOneReceived(echoic=echoic)


    def serviceReceives(self, *, echoic=False):
        """Service all receives (greedy) and queue up

        Parameters:
            echoic (bool): True means use .echos in .receive debugging purposes
                            where echo is duple of form: (gram: bytes, src: str)
                           False means do not use .echos default is duple that
                            indicates nothing to receive of form (b'', None)
        """
        while self.opened:
            if not self._serviceOneReceived(echoic=echoic):
                break


    def fuse(self, grams, cnt):
        """Fuse cnt gram body parts from grams dict into whole memo . If any
        grams are missing then returns None.

        Returns:
            memo (str | None): fused memo or None if incomplete.

        Override in subclass

        Parameters:
            grams (dict): memo gram body parts each keyed by gram number from which
                          to fuse memo. Headers have been stripped.
            cnt (int): gram count for mid
        """
        if len(grams) < cnt:  # must be missing one or more grams
            return None

        memo = bytearray()
        for i in range(cnt):  # iterate in numeric order, items are insertion ordered
            memo.extend(grams[i])  # extend memo with gram body part at gram i

        return memo.decode()  # convert bytearray to str



    def _serviceOnceRxGrams(self):
        """Service one pass over .rxgs dict for each unique mid in .rxgs

        Deleting an item from a dict at a key (since python dicts are key
        insertion ordered) means that the next time an item is created it will
        be last.
        """
        for i, mid in enumerate(list(self.rxgs.keys())):  # items may be deleted in loop
            # if mid then grams dict at mid must not be empty
            if not mid in self.counts:  # missing first gram so skip
                continue
            memo = self.fuse(self.rxgs[mid], self.counts[mid])
            if memo is not None:  # allows for empty "" memo for some src
                self.rxms.append((memo, self.sources[mid], self.vids[mid]))
                del self.rxgs[mid]
                del self.counts[mid]
                del self.sources[mid]
                del self.vids[mid]


    def serviceRxGramsOnce(self):
        """Service one pass (non-greedy) over all unique sources in .rxgs
        dict if any for received incoming grams.
        """
        if self.rxgs:
            self._serviceOnceRxGrams()


    def serviceRxGrams(self):
        """Service one pass (non-greedy) over all unique sources in .rxgs
        dict if any for received incoming grams.  No different from
        serviceRxGramsOnce because service all mids each pass.
        """
        if self.rxgs:
            self._serviceOnceRxGrams()


    def _serviceOneRxMemo(self):
        """Service one duple from .rxMsgs deque

        Returns:
            duple (tuple | None): of form (memo: str, src: str) if any
                                  otherwise None

        """
        return (self.rxms.popleft())  # raises IndexError if empty deque


    def serviceRxMemosOnce(self):
        """Service memos in .rxms deque once (non-greedy one memo) if any

        Override in subclass to handle result and put it somewhere
        """
        try:
            memo, src, vid = self._serviceOneRxMemo()
        except IndexError:
            pass


    def serviceRxMemos(self):
        """Service all memos in .rxms (greedy) if any

        Override in subclass to handle result(s) and put them somewhere
        """
        while self.rxms:
            memo, src, vid = self._serviceOneRxMemo()


    def serviceAllRxOnce(self):
        """Service receive side of stack once (non-greedy)
        """
        self.serviceReceivesOnce()
        self.serviceRxGramsOnce()
        self.serviceRxMemosOnce()


    def serviceAllRx(self):
        """Service receive side of stack (greedy) until empty or waiting for more
        """
        self.serviceReceives()
        self.serviceRxGrams()
        self.serviceRxMemos()


    def send(self, txbs, dst, *, echoic=False):
        """Attemps to send bytes in txbs to remote destination dst.
        Must be overridden in subclass.
        This is a stub to define mixin interface.

        Returns:
            count (int): bytes actually sent

        Parameters:
            txbs (bytes | bytearray): of bytes to send
            dst (str): remote destination address
            echoic (bool): True means echo sends into receives via. echos
                           False measn do not echo
        """
        if echoic:
            self.echos.append((bytes(txbs), dst))  # make copy
        return len(txbs)  # all sent



    def memoit(self, memo, dst, vid=None):
        """Append (memo, dst, vid) tuple to .txms deque

        Parameters:
            memo (str): to be segmented and packed into gram(s)
            dst (str): address of remote destination of memo
            vid (str | None): verifiable ID for signing grams
        """
        self.txms.append((memo, dst, vid))


    def sign(self, ser, vid):
        """Sign serialization ser using private key for verifier ID vid
        Must be overriden in subclass to fetch private key for vid and sign.
        This is a stub.

        Returns:
            sig(bytes): qb64b qualified base64 representation of signature

        Parameters:
            ser (bytes): signed portion of gram is delivered format
            vid (str | bytes): qualified base64 qb64 of verifier ID

        """
        if hasattr(vid, 'encode'):
            vid = vid.encode()
        cs, ms, vs, ss, ns, hs = self.Sizes[self.code]  # cs ms vs ss ns hs
        return b'A' * ss


    def rend(self, memo, vid=None):
        """Partition memo into packed grams with headers.

        Returns:
            grams (list[bytes]): list of grams with headers.

        Parameters:
            memo (str): to be partitioned into grams with headers
            vid (str | None): verification ID when gram is to be signed.
                              None means not signable

        Note first gram has head + neck overhead, hs + ns so bs is smaller by ns
             non-first grams have just head overhead hs so bs is bigger by ns
        """
        grams = []
        memo = bytearray(memo.encode()) # convert and copy to bytearray
        # self.size is max gram size
        cs, ms, vs, ss, ns, hs = self.Sizes[self.code]  # cs ms vs ss ns hs
        ps = (3 - ((ms) % 3)) % 3  # net pad size for mid
        if vs and (not vid or len(vid) != vs):
            raise hioing.MemoerError(f"Invalid {vid=} for size={vs}.")

        vid = b"" if vid is None else vid[:vs].encode()

        # memo ID is 16 byte random UUID converted to 22 char Base64 right aligned
        mid = encodeB64(bytes([0] * ps) + uuid.uuid4().bytes)[ps:] # prepad, convert, and strip
        mid = self.code.encode() + mid  # fully qualified mid with prefix code
        ml = len(memo)

        if self.curt:  # rend header parts in base2 instead of base64
            hs = 3 * hs // 4  # encoding b2 means head part sizes smaller by 3/4
            ns = 3 * ns // 4  # encoding b2 means head part sizes smaller by 3/4
            vs = 3 * vs // 4  # encoding b2 means head part sizes smaller by 3/4
            ss = 3 * ss // 4  # encoding b2 means head part sizes smaller by 3/4
            mid = decodeB64(mid)
            vid = decodeB64(vid)

        bs = (self.size - hs)  # max standard gram body size without neck
        # compute gram count based on overhead note added neck overhead in first gram
        # first gram is special its header is longer by ns than the other grams
        # which means its payload body is shorter by ns than the other gram bodies
        # so take ml and subtract first payload size = ml - (bs-ns) to get the
        # portion of memo carried by other grams. Divide this by bs rounded up
        # (ceil) to get cnt of other grams and add 1 for the first gram to get
        # total gram cnt.
        # gc = ceil((ml-(bs-ns))/bs + 1) = ceil((ml-bs+ns)/bs + 1)
        gc = math.ceil((ml-bs+ns)/bs+1)  # includes added neck ns overhead
        mms = min(self.MaxMemoSize, (bs * self.MaxGramCount) - ns)  # max memo payload

        if ml > mms:
            raise hioing.MemoerError(f"Memo length={ml} exceeds max={mms}.")

        if self.curt:
            neck = gc.to_bytes(ns)
        else:
            neck = helping.intToB64b(gc, l=ns)

        gn = 0
        while memo:
            if self.curt:
                num = gn.to_bytes(ns)  # num size must always be neck size
            else:
                num = helping.intToB64b(gn, l=ns)  # num size must always be neck size

            head = mid + vid + num

            if gn == 0:
                gram = head + neck + memo[:bs-ns]  # copy slice past end just copies to end
                del memo[:bs-ns]  # del slice past end just deletes to end
            else:
                gram = head + memo[:bs]  # copy slice past end just copies to end
                del memo[:bs]  # del slice past end just deletes to end

            if ss:  # sign
                sig = self.sign(gram, vid)
                if self.curt:
                    sig = decodeB64(sig)
                gram = gram + sig

            grams.append(gram)

            gn += 1

        return grams


    def _serviceOneTxMemo(self):
        """Service one (memo, dst, vid) tuple from .txms deque where tuple is
        of form: (memo: str, dst: str, vid: str) where:
                memo is outgoing memo
                dst is destination address
                vid is verification ID

        Calls .rend method to process the partitioning and packing as
        appropriate to convert memo into grams with headers and sign when
        indicated.

        Appends (gram, dst) duple to .txgs deque.
        """
        memo, dst, vid = self.txms.popleft()  # raises IndexError if empty deque

        for gram in self.rend(memo, vid):  # partition memo into gram parts with head
            self.txgs.append((gram, dst))  # append duples (gram: bytes, dst: str)


    def serviceTxMemosOnce(self):
        """Service one outgoing memo from .txms deque if any (non-greedy)
        """

        try:
            self._serviceOneTxMemo()
        except IndexError:
            pass


    def serviceTxMemos(self):
        """Service all outgoing memos in .txms deque if any (greedy)
        """
        while self.txms:
            self._serviceOneTxMemo()


    def gramit(self, gram, dst):
        """Append (gram, dst) duple to .txgs deque

        Parameters:
            gram (bytes): gram to be sent
            dst (str): address of remote destination of gram
        """
        self.txgs.append((gram, dst))


    def _serviceOnceTxGrams(self, *, echoic=False):
        """Service one pass over .txgs dict for each unique dst in .txgs and then
        services one pass over the .txbs of those outgoing grams

        Parameters:
           echoic (bool): True means echo sends into receives via. echos
                           False measn do not echo

        Each entry in.txgs is duple of form: (gram: bytes, dst: str)
        where:
            gram (bytes): is outgoing gram segment from associated memo
            dst (str): is far peer destination address

        .txbs is duple of form: (gram: bytearray, dst: str)
        where:
            gram (bytearray): holds incompletly sent gram portion if any
            dst (str | None): destination or None if last completely sent

        Returns:
            result (bool):  True means greedy callers can keep sending since
                                last gram sent completely.
                            False means either last gram send was incomplete and
                                or there are no more grams in .txgs deque. In
                                either case greedy callers need to wait and
                                try again later.

        The return value True or False enables back pressure on greedy callers
        so they know when to block waiting.

        When there is a remainder in .txbs each subsequent call of this method
        will attempt to send the remainder until the the full gram has been sent.
        This accounts for datagram protocols that expect continuing attempts to
        send remainder of a datagram when using nonblocking sends.

        When the far side peer is unavailable the gram is dropped. This means
        that unreliable transports need to have a timeour retry mechanism.

        Internally, a dst of None in the .txbs duple indicates its ok to take
        another gram from the .txgs deque if any and start sending it.

        """
        gram, dst = self.txbs
        if dst == None:
            try:
                gram, dst = self.txgs.popleft()
                gram = bytearray(gram)
            except IndexError:
                return False  # nothing more to send, return False to try later

        cnt = 0
        try:
            cnt = self.send(gram, dst, echoic=echoic)  # assumes .opened == True
        except socket.error as ex:  # OSError.errno always .args[0] for compat
            if (ex.args[0] in (errno.ECONNREFUSED,
                               errno.ENOENT,
                               errno.ECONNRESET,
                               errno.ENETRESET,
                               errno.ENETUNREACH,
                               errno.EHOSTUNREACH,
                               errno.ENETDOWN,
                               errno.EHOSTDOWN,
                               errno.ETIMEDOUT,
                               errno.ETIME)):  # far peer problem
                # try again later usually won't work here so we log error
                # and drop gram so as to allow grams to other destinations
                # to get sent. When uxd, ECONNREFUSED and ENOENT means dest
                # uxd file path is not available to send to.
                logger.error("Error send from %s to %s\n %s\n",
                                                         self.name, dst, ex)
                self.txbs = (bytearray(), None) # far peer unavailable, so drop.
                dst = None  # dropped is same as all sent
            else:
                raise  # unexpected error

        if cnt:
            del gram[:cnt]  # remove from buffer those bytes sent
            if not gram:  # all sent
                dst = None  # indicate by setting dst to None
            self.txbs = (gram, dst)  # update txbs to indicate if completely sent

        return (False if dst else True)  # incomplete return False, else True


    def serviceTxGramsOnce(self, *, echoic=False):
        """Service one pass (non-greedy) over all unique destinations in .txgs
        dict if any for blocked destination or unblocked with pending outgoing
        grams.

        Parameters:
           echoic (bool): True means echo sends into receives via. echos
                           False measn do not echo
        """
        if self.opened and self.txgs:
            self._serviceOnceTxGrams(echoic=echoic)


    def serviceTxGrams(self, *, echoic=False):
        """Service multiple passes (greedy) over all unqique destinations in
        .txgs dict if any for blocked destinations or unblocked with pending
        outgoing grams until there is no unblocked destination with a pending gram.

        Parameters:
           echoic (bool): True means echo sends into receives via. echos
                           False measn do not echo
        """
        while self.opened and self.txgs:  # pending gram(s)
            if not self._serviceOnceTxGrams(echoic=echoic):  # send incomplete
                break  # try again later


    def serviceAllTxOnce(self):
        """Service the transmit side once (non-greedy) one transmission.
        """
        self.serviceTxMemosOnce()
        self.serviceTxGramsOnce()


    def serviceAllTx(self):
        """Service the transmit side until empty (greedy) multiple transmissions
        until blocked by pending transmission on transport.
        """
        self.serviceTxMemos()
        self.serviceTxGrams()


    def serviceLocal(self):
        """Service the local Peer's receive and transmit queues
        """
        self.serviceReceives()
        self.serviceTxGrams()


    def serviceAllOnce(self):
        """Service all Rx and Tx Once (non-greedy)
        """
        self.serviceAllRxOnce()
        self.serviceAllTxOnce()


    def serviceAll(self):
        """Service all Rx and Tx (greedy)
        """
        self.serviceAllRx()
        self.serviceAllTx()

    service = serviceAll  # alias


@contextmanager
def openMemoer(cls=None, name="test", **kwa):
    """
    Wrapper to create and open Memoer instances
    When used in with statement block, calls .close() on exit of with block

    Parameters:
        cls (Class): instance of subclass instance
        name (str): unique identifer of Memoer peer.
            Enables management of transport by name.
    Usage:
        with openMemoer() as peer:
            peer.receive()

        with openMemoer(cls=MemoerSub) as peer:
            peer.receive()

    """
    peer = None

    if cls is None:
        cls = Memoer
    try:
        peer = cls(name=name, **kwa)
        peer.reopen()

        yield peer

    finally:
        if peer:
            peer.close()


class MemoerDoer(doing.Doer):
    """Memoer Doer for reliable transports that do not require retry tymers.

    See Doer for inherited attributes, properties, and methods.

    Attributes:
       .peer (Memoer): underlying transport instance subclass of Memoer

    """

    def __init__(self, peer, **kwa):
        """Initialize instance.

        Parameters:
           peer (Peer): is Memoer Subclass instance
        """
        super(MemoerDoer, self).__init__(**kwa)
        self.peer = peer


    def enter(self, *, temp=None):
        """Do 'enter' context actions. Override in subclass. Not a generator method.
        Set up resources. Comparable to context manager enter.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any
        """
        # inject temp into file resources here if any
        self.peer.reopen(temp=temp)


    def recur(self, tyme):
        """"""
        self.peer.service()


    def exit(self):
        """"""
        self.peer.close()



class TymeeMemoer(tyming.Tymee, Memoer):
    """TymeeMemoer mixin base class to add tymer support for unreliable transports
    that need retry tymers. Subclass of tyming.Tymee


    Inherited Class Attributes:
        see superclass

    Class Attributes:
        Tymeout (float): default timeout for retry tymer(s) if any

    Inherited Attributes:
        see superclass

    Attributes:
        tymeout (float): default timeout for retry tymer(s) if any



    """
    Tymeout = 0.0  # tymeout in seconds, tymeout of 0.0 means ignore tymeout

    def __init__(self, *, tymeout=None, **kwa):
        """
        Initialization method for instance.
        Inherited Parameters:
            see superclass

        Parameters:
            tymeout (float): default for retry tymer if any

        """
        super(TymeeMemoer, self).__init__(**kwa)
        self.tymeout = tymeout if tymeout is not None else self.Tymeout
        #self.tymer = tyming.Tymer(tymth=self.tymth, duration=self.tymeout) # retry tymer


    def wind(self, tymth):
        """
        Inject new tymist.tymth as new ._tymth. Changes tymist.tyme base.
        Updates winds .tymer .tymth
        """
        super(TymeeMemoer, self).wind(tymth)
        #self.tymer.wind(tymth)

    def serviceTymers(self):
        """Service all retry tymers

        Must be overriden in subclass.
        This is a stub.
        """
        pass


    def serviceLocal(self):
        """Service the local Peer's receive and transmit queues
        """
        self.serviceReceives()
        self.serviceTxGrams()


    def serviceAllOnce(self):
        """Service all Rx and Tx Once (non-greedy)
        """
        self.serviceAllRxOnce()
        self.serviceTymers()
        self.serviceAllTxOnce()


    def serviceAll(self):
        """Service all Rx and Tx (greedy)
        """
        self.serviceAllRx()
        self.serviceTymers()
        self.serviceAllTx()


@contextmanager
def openTM(cls=None, name="test", **kwa):
    """
    Wrapper to create and open TymeeMemoer instances
    When used in with statement block, calls .close() on exit of with block

    Parameters:
        cls (Class): instance of subclass instance
        name (str): unique identifer of Memoer peer.
            Enables management of transport by name.
    Usage:
        with openTM() as peer:
            peer.receive()

        with openTM(cls=MemoerSub) as peer:
            peer.receive()

    """
    peer = None

    if cls is None:
        cls = TymeeMemoer
    try:
        peer = cls(name=name, **kwa)
        peer.reopen()

        yield peer

    finally:
        if peer:
            peer.close()


class TymeeMemoerDoer(doing.Doer):
    """TymeeMemoer Doer for unreliable transports that require retry tymers.

    See Doer for inherited attributes, properties, and methods.

    Attributes:
       .peer (TymeeMemoer) is underlying transport instance subclass of TymeeMemoer

    """

    def __init__(self, peer, **kwa):
        """Initialize instance.

        Parameters:
           peer (TymeeMemoer):  subclass instance
        """
        super(TymeeMemoerDoer, self).__init__(**kwa)
        self.peer = peer



    def wind(self, tymth):
        """Inject new tymist.tymth as new ._tymth. Changes tymist.tyme base.
        Updates winds .tymer .tymth
        """
        super(TymeeMemoerDoer, self).wind(tymth)
        self.peer.wind(tymth)


    def enter(self, *, temp=None):
        """Do 'enter' context actions. Override in subclass. Not a generator method.
        Set up resources. Comparable to context manager enter.

        Parameters:
            temp (bool | None): True means use temporary file resources if any
                                None means ignore parameter value. Use self.temp

        Inject temp or self.temp into file resources here if any

        Doist or DoDoer winds its doers on enter
        """
        # inject temp into file resources here if any
        if self.tymth:
            self.peer.wind(self.tymth)
        self.peer.reopen(temp=temp)


    def recur(self, tyme):
        """"""
        self.peer.service()


    def exit(self):
        """"""
        self.peer.close()




# -*- encoding: utf-8 -*-
"""hio.help.naming Module

Mixin Base Class for one-to-one mapping with inverse: name-to-addr and addr-to-name
that add support for MemoGrams to datagram based transports.

Provide mixin classes.
Takes of advantage of multiple inheritance to enable mixtures of behaviors
with minimal code duplication (more DRY).

New style python classes use the C3 linearization algorithm. Multiple inheritance
forms a directed acyclic graph called a diamond graph. This graph is linarized
into the method resolution order.
Use class.mro() or class.__mro__

(see https://www.geeksforgeeks.org/method-resolution-order-in-python-inheritance/)
Basically:
* children always precede their parents
* immediate parent classes of a child are visited in the order listed in the
child class statement.
* a super class is visited only after all sub classes have been visited
* linearized graph is monotonic (a class is only visted once)

"""


from .. import hioing
from . import helping

class Namer(hioing.Mixin):
    """
    Namer mixin class for adding support for mappings between names and addresses.
    May be used to provide in-memory lookup of mapping and its inverse of
    address from name and name from address.

    """

    def __init__(self, entries=None, **kwa):
        """
        Setup instance

        Inherited Parameters:

        Parameters:
            entries (dict | interable): dict or iterable of (name, addr) duples
                to  bulk initialize the mappings between addresses and names

        Attributes:

        Properties:
            addrByName (dict): mapping between (name, address) pairs, these
                must be one-to-one so that inverse is also one-to-one
            nameByAddr (dict): mapping between (address, name) pairs, these
                must be one-to-one so that inverse is also one-to-one

        Hidden:
            _addrByName (dict): mapping between (name, address) pairs, these
                must be one-to-one so that inverse is also one-to-one
            _nameByAddr (dict): mapping between (address, name) pairs, these
                must be one-to-one so that inverse is also one-to-one

        Methods:
            clearAllNameAddr()
            getAddr(name)
            getName(addr)
            addNameAddr(name, addr)
            remNameAddr(name=None, addr=None)
            changeAddrAtName(name=None, addr=None)
            changeNameAtAddr(addr=None, name=None)

        """
        self._addrByName = dict()
        self._nameByAddr = dict()

        if helping.isNonStringIterable(entries):
            items = entries
        else:
            items = entries.items() if entries else []

        for name, addr in items:
            self.addNameAddr(name=name, addr=addr)

        super(Namer, self).__init__(**kwa)


    @property
    def addrByName(self):
        """Property that returns copy of ._addrByName
        """
        return dict(self._addrByName)

    @property
    def nameByAddr(self):
        """Property that returns copy of ._nameByAddr
        """
        return dict(self._nameByAddr)

    @property
    def countNameAddr(self):
        """Property that returns count of entries in .addrByName
        """
        return len(self._addrByName)


    def clearAllNameAddr(self):
        """Clears all entries

        """
        self._addrByName = dict()
        self._nameByAddr = dict()


    def getAddr(self, name):
        """Returns addr for given name or None if non-existant

        Parameters:
            name (str): name
        """
        return self._addrByName.get(name)


    def getName(self, addr):
        """Returns name for given addr or None if non-existant

        Parameters:
            addr (str): address
        """
        return self._nameByAddr.get(addr)


    def addNameAddr(self, name, addr):
        """Add mapping and inverse mapping entries for
        name to addr and addr to name
        All mappings must be one-to-one

        Returns:
            result (bool):  True if successfully added new entry.
                            False If matching entry already exists.
                            Otherwise raises NamerError if partial matching
                                entry with either name or addr

        Parameters:
            name (str): name of entry
            addr (str): address of entry

        Raise error if preexistant
        """
        if not name or not addr:
            raise  hioing.NamerError(f"Attempt to add incomplete or empty entry "
                                   f"({name=}, {addr=}).")
        if name in self._addrByName:
            if addr == self._addrByName[name]:
                return False  # already existing matching entry
            else:
                raise  hioing.NamerError(f"Attempt to add conflicting entry "
                               f"({name=}, {addr=}).")

        if addr in self._nameByAddr:
            if name == self._nameByAddr[addr]:
                return False  # already existing matching entry
            else:
                raise  hioing.NamerError(f"Attempt to add conflicting entry "
                               f"({name=}, {addr=}).")

        self._addrByName[name] = addr
        self._nameByAddr[addr] = name

        return True  # added new entry


    def remNameAddr(self, name=None, addr=None):
        """Delete both name to addr and addr to name mapping entries.
           When an entry is found for either name or addr or both.
           When both provided must be matching one-to-one entries.
           All mappings must be one-to-one

        Returns:
            result (bool):  True if matching entry successfully deleted
                            False if no matching entry, nothing deleted

        Parameters:
            name (str | None): name of remote to delete or None if delete by addr
            addr (str | None): addr of remote to delete or None if delete by name

            When both name and addr provided deletes by name

        Raise error when at least one of name or addr is not None and no mappings
        exist.
        """

        if name:
            if name not in self._addrByName:  # no entry for name
                return False

            if not addr:  # addr not provided so look it up
                addr = self._addrByName[name]

            if addr != self._addrByName[name]:  # mismatch do nothing
                return False

            del self._addrByName[name]
            del self._nameByAddr[addr]
            return True

        if addr:
            if addr not in self._nameByAddr:  # no entry for addr
                return False

            if not name:  # no name so look it up
                name = self._nameByAddr[addr]

            if name != self._nameByAddr[addr]:  # mismatch do nothing
                return False

            del self._addrByName[name]
            del self._nameByAddr[addr]
            return True

        return False  # nothing deleted neither name nor addr provided



    def changeAddrAtName(self, *, name=None, addr=None):
        """Changes existing name to addr mapping to new addr. Replaces
        inverse mapping of addr to name.
        All mappings must be one-to-one

        Returns:
            result (bool):  True If successfully updated existing entries.
                            False If matching entry already exists, no change.
                            Otherwise raises NamerError

        Parameters:
            name (str): name of entry
            addr (str): address of entry

        Raise error if preexistant
        """
        if not name or not addr:
            raise  hioing.NamerError(f"Attempt to update with incomplete "
                                     f"or empty entry ({name=}, {addr=}).")

        if name not in self._addrByName:
            return False

        if addr == self._addrByName[name]:
            return False  # no change to entries

        if addr in self._nameByAddr:
            raise  hioing.NamerError(f"Conflicting entry for {addr=}.")

        oldAddr = self._addrByName[name]
        self._addrByName[name] = addr
        del self._nameByAddr[oldAddr]
        self._nameByAddr[addr] = name

        return True  # updated entries


    def changeNameAtAddr(self, *, addr=None, name=None):
        """Changes existing addr to name mapping to new name. Replaces
        inverse mapping of name to addr.
        All mappings must be one-to-one

        Returns:
            result (bool):  True If successfully updated existing entries.
                            False If matching entry already exists, no change.
                            Otherwise raises NamerError

        Parameters:
            addr (str): address of entry
            name (str): name of entry

        Raise error if preexistant
        """
        if not name or not addr:
            raise  hioing.NamerError(f"Attempt to update with incomplete entry "
                                   f"({name=}, {addr=}).")

        if addr not in self._nameByAddr:
            return False

        if name == self._nameByAddr[addr]:
            return False  # no change to entries

        if name in self._addrByName:
            raise  hioing.NamerError(f"Conflicting entry for {name=}.")

        oldName = self._nameByAddr[addr]
        self._nameByAddr[addr] = name
        del self._addrByName[oldName]
        self._addrByName[name] = addr

        return True  # updated entries




# -*- encoding: utf-8 -*-
"""
hio.help.helping module

"""
from __future__ import annotations  # so type hints of classes get resolved later

import types
import functools
import inspect
import os
import errno
import stat

import base64
import re

import json
import msgpack
import cbor2 as cbor

from typing import Any, Type
from collections import deque
from collections.abc import Iterable, Sequence, Generator, Mapping, Callable
from abc import ABCMeta
from dataclasses import dataclass, astuple, asdict, fields, field




# Utilities
def isign(i):
    """
    Integer sign function
    Returns:
        (int): 1 if i > 0, -1 if i < 0, 0 otherwise

    """
    return (1 if i > 0 else -1 if i < 0 else 0)


def sceil(r):
    """
    Symmetric ceiling function
    Returns:
       sceil (int): value that is symmetric ceiling of r away from zero

    Because int() provides a symmetric floor towards zero, just inc int(r) by:
     1 when r - int(r) >  0  (r positive)
    -1 when r - int(r) <  0  (r negative)
     0 when r - int(r) == 0  (r integral already)
    abs(r) > abs(int(r) or 0 when abs(r)
    """
    return (int(r) + isign(r - int(r)))


def copyfunc(f, name=None):
    """
    Copy a function in detail.
    To change name of func provide name parameter

    functools to update_wrapper assigns and updates following attributes
    WRAPPER_ASSIGNMENTS = ('__module__', '__name__', '__qualname__', '__doc__',
                       '__annotations__')
    WRAPPER_UPDATES = ('__dict__',)
    Based on
    https://stackoverflow.com/questions/6527633/how-can-i-make-a-deepcopy-of-a-function-in-python
    https://stackoverflow.com/questions/13503079/how-to-create-a-copy-of-a-python-function
    """
    g = types.FunctionType(f.__code__,
                           f.__globals__,
                           name=f.__name__,
                           argdefs=f.__defaults__,
                           closure=f.__closure__
                          )
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    if name:
        g.__name__ = name
    return g


def attributize(genie):
    """
    Decorator function:

    Python generators do not support adding attributes.
    Adding support for attributes provides a way to pass information
    from a WSGI App that returns a generator to a WSGI server via the generator
    after the WSGI app has already started returning its body. The hio.http.Server
    WSGI server looks for the attributes ._status and ._headers and substitutes
    these if present. This allows a streaming WSGI App body iterator to later
    modify the headers and status taht will be returned before the body iterator
    began iterating. This is useful for web hooks or backend requests that are
    serviced by an async coroutine based WSGI app so that they may leverage
    the streaming support of standard WSGI but use a the coroutine based
    hio.http.Server as an async WSGI server.

    This decorator takes a Duck Typing approach to decorating
    a generator function or method that returns a new function type instance that
    when called will return a generator like object that supports attributes.
    the new wrapped object acts like a generator but with attributes.

    Parameters:
        genie (generator function, generator method): is either
            a generator function that returns a generator object
            a generator method that returns a generator object

    If genie is a generator function then a reference to its wrapper
         is injected as the first positional argument to the orginal
         generator function. The convention is to use the parameter 'me' to refer
         to the injected reference to the wrapper.

    If genie is a generator method, that is, its first parameter is 'self'
        then a reference to its wrapper is injected as the second
        positional argument to the original generator method. the convention
        is to use the parameter 'me' to refer to the injected reference to the
        wrapper so as not to collide with the 'self' instance reference.

    When wrapped the new type is AttributiveGenerator

    Usage:
    # decorated generator function
    @attributize
    def bar(me, req=None, rep=None):
        me._status = 400  # or copy from rep.status
        me._headers = odict(example="Hi")  # or copy from rep.headers
        yield b""
        yield b""
        yield b"Hello There"
        return b"Goodbye"

    gen = bar()
    msg = next(gen)  # attributes set after first next
    gen._status
    gen._headers

    # decorated generator method
    class R:
        @attributize
        def bar(self, me, req=None, rep=None):
            self.name = "Peter"
            me._status = 400  # or copy from rep.status
            me._headers = odict(example="Hi")  # or copy from rep.headers
            yield b""
            yield b""
            yield b"Hello There " + self.name.encode()
            return b"Goodbye"

    r = R()
    gen = r.bar()
    msg = next(gen)   # attributes set after first next
    gen._status
    gen._headers

    # use as function wrapper directly instead of as decorator
    def gf(me, x):  # convention injected reference to attributed wrapper is 'me'
        me.x = 5
        me.y = 'a'
        cnt = 0
        while cnt < x:
            yield cnt
            cnt += 1

    agf = attributize(gf)
    ag = agf(3)
    # body of gf is not run until first next call
    assert isIterator(ag)
    assert not hasattr(ag, 'x')
    assert not hasattr(ag, 'y')
    n = next(ag)  # first run here which sets up attributes
    assert n == 0
    assert hasattr(ag, 'x')
    assert hasattr(ag, 'y')
    assert ag.x == 5
    assert ag.y =='a'
    n = next(ag)
    assert n == 1

    Adding attributes to this injected reference makes them available
    as attributes of the resultant wrapper.

    The HTTP WSGI server at hio.core.http.serving.Server uses an instance of
    hio.core.http.serving.Responder to generate the response for each WSGI request.
    The Responder instance checks its WSGI app generator for existence of the
    attributes ._status and ._headers. If so then it overrides its default
    response status with ._status and updates its default response headers
    with the headers in ._header. This allows a backend (webhook) to conveniently
    influence the response status and headers. The response body is returned by
    the generator itself.

    Background:
    Unlike Python functions, Python generators do not support  custom attributes
    and the generator locals dict at .gi_frame.f_locals dissappears once the
    generator is complete so its inconvenient.

    Fixed attributes of generator objects.
    ['.__next__', '__iter__', 'close', 'gi_code', 'gi_frame', 'gi_running',
    'gi_yieldfrom', 'send', 'throw']
    """

    def wrapper(*args, **kwargs):
        """
        When called returns instance of AttributiveGenerator instead of generator.
        """
        def __iter__(self):  # default attribute
            return self

        def send(self):   # default attribute
            raise NotImplementedError

        def throw(self):   # default attribute
            raise NotImplementedError

        # use type() to create dynamic instance of class AttributiveGenerator
        # dict spec for type() is {'__iter__': lambda self: self, 'send': ,}
        tdict = { '__iter__': __iter__, 'send': send, 'throw':  throw,}
        AG = type("AttributiveGenerator", (Generator,), tdict)  # make custom type
        ag = AG()  # create  instance so we can inject it into genfunc

        # create generator and inject ag ref for me parameter  ag = me
        fargs = inspect.getfullargspec(genie).args
        if fargs and fargs[0] == 'self':
            gen = genie(args[0], ag, *args[1:], **kwargs)  # inject self and me
        else:
            gen = genie(ag, *args, **kwargs)  # inject me only

        # now replace default class references with real gen attributes. "duckify"
        for attr in ('__next__', 'close', 'send', 'throw',
                     'gi_code', 'gi_frame', 'gi_running', 'gi_yieldfrom'):
            setattr(AG, attr, getattr(gen, attr))  # methods are class attributes

        functools.update_wrapper(wrapper=ag, wrapped=gen)  # fix up wrapper
        return ag

    return wrapper


def repack(n, seq, default=None):
    """ Repacks seq into a generator of len n and returns the generator.
        The purpose is to enable unpacking into n variables.
        The first n-1 elements of seq are returned as the first n-1 elements of the
        generator and any remaining elements are returned in a tuple as the
        last element of the generator
        default (None) is substituted for missing elements when len(seq) < n

        Example:

        x = (1, 2, 3, 4)
        tuple(repack(3, x))
        (1, 2, (3, 4))

        x = (1, 2, 3)
        tuple(repack(3, x))
        (1, 2, (3,))

        x = (1, 2)
        tuple(repack(3, x))
        (1, 2, ())

        x = (1, )
        tuple(repack(3, x))
        (1, None, ())

        x = ()
        tuple(repack(3, x))
        (None, None, ())

    """
    it = iter(seq)
    for i in range(n - 1):
        yield next(it, default)
    yield tuple(it)


def just(n, seq, default=None):
    """ Returns a generator of just the first n elements of seq and substitutes
        default (None) for any missing elements. This guarantees that a generator of exactly
        n elements is returned. This is to enable unpacking into n varaibles

        Example:

        x = (1, 2, 3, 4)
        tuple(just(3, x))
        (1, 2, 3)
        x = (1, 2, 3)
        tuple(just(3, x))
        (1, 2, 3)
        x = (1, 2)
        tuple(just(3, x))
        (1, 2, None)
        x = (1, )
        tuple(just(3, x))
        (1, None, None)
        x = ()
        tuple(just(3, x))
        (None, None, None)

    """
    it = iter(seq)
    for i in range(n):
        yield next(it, default)


class NonStringIterable(metaclass=ABCMeta):
    """
    Allows isinstance check for iterable that is not a string
    if isinstance(x, NonStringIterable):
    """
    @classmethod
    def __subclasshook__(cls, C):
        if cls is NonStringIterable:
            if (not issubclass(C, (str, bytes)) and issubclass(C, Iterable)):
                return True
        return NotImplemented


class NonStringSequence(metaclass=ABCMeta):
    """
    Allows isinstance check for sequence that is not a string
    if isinstance(x, NonStringSequence):
    """
    @classmethod
    def __subclasshook__(cls, C):
        if cls is NonStringSequence:
            if (not issubclass(C, (str, bytes)) and issubclass(C, Sequence)):
                return True
        return NotImplemented


def isNonStringIterable(obj):
    """
    Returns: (bool) True if obj is non-string iterable, False otherwise

    Another way that is less future proof
    return (hasattr(x, '__iter__') and not isinstance(x, (str, bytes)))
    """
    return (not isinstance(obj, (str, bytes)) and isinstance(obj, Iterable))

nonStringIterable = isNonStringIterable  # legacy interface


def isNonStringSequence(obj):
    """
    Returns: (bool) True if obj is non-string sequence, False otherwise
    """
    return (not isinstance(obj, (str, bytes)) and isinstance(obj, Sequence) )

nonStringSequence = isNonStringSequence  # legacy interface


def isIterator(obj):
    """
    Returns True if obj is an iterator object, that is,

    has an __iter__ method
    has a __next__ method
    .__iter__ is callable and returns obj

    Otherwise returns False

    """
    if (hasattr(obj, "__iter__") and
        hasattr(obj, "__next__") and
        callable(obj.__iter__) and
        obj.__iter__() is obj
       ):
        return True
    return False


def ocfn(path, mode='r+', perm=(stat.S_IRUSR | stat.S_IWUSR)):
    """
    Atomically open or create file from filepath.

    If file already exists, Then open file using openMode
    Else create file using write update mode If not binary Else
        write update binary mode
    Returns file object

    If binary Then If new file open with write update binary mode

    x = stat.S_IRUSR | stat.S_IWUSR
    384 == 0o600
    436 == octal 0664
    """
    try:

        newfd = os.open(path, os.O_EXCL | os.O_CREAT | os.O_RDWR, perm)
        if "b" in mode:
            file = os.fdopen(newfd,"w+b")  # w+ truncate read and/or write
        else:
            file = os.fdopen(newfd,"w+")  # w+ truncate read and/or write

    except OSError as ex:
        if ex.errno == errno.EEXIST:
            file = open(path, mode)  # r+ do not truncate read and/or write
        else:
            raise
    return file


def dump(data, path):
    """
    Serialize data dict and write to file given by path where serialization is
    given by path's extension of either JSON, MsgPack, or CBOR for extension
    .json, .mgpk, or .cbor respectively
    """

    if ' ' in path:
        raise IOError(f"Invalid file path '{path}' contains space.")

    root, ext = os.path.splitext(path)
    if ext == '.json':
        with ocfn(path, "w+b") as f:
            json.dump(data, f, indent=2)
            f.flush()
            os.fsync(f.fileno())
    elif ext == '.mgpk':
        with ocfn(path, "w+b") as f:
            msgpack.dump(data, f)
            f.flush()
            os.fsync(f.fileno())
    elif ext == '.cbor':
        with ocfn(path, "w+b") as f:
            cbor.dump(data, f)
            f.flush()
            os.fsync(f.fileno())
    else:
        raise IOError(f"Invalid file path ext '{path}' "
                      f"not '.json', '.mgpk', or 'cbor'.")


def load(path):
    """
    Return data read from file path as dict
    file may be either json, msgpack, or cbor given by extension .json, .mgpk, or
    .cbor respectively
    Otherwise raise IOError
    """
    root, ext = os.path.splitext(path)
    if ext == '.json':
        with ocfn(path, "rb") as f:
            it = json.load(f)
    elif ext == '.mgpk':
        with ocfn(path, "rb") as f:
            it = msgpack.load(f)
    elif ext == '.cbor':
        with ocfn(path, "rb") as f:
            it = cbor.load(f)
    else:
        raise IOError(f"Invalid file path ext '{path}' "
                     f"not '.json', '.mgpk', or 'cbor'.")

    return it


# Base64 utilities

# Mappings between Base64 Encode Index and Decode Characters
#  B64ChrByIdx is dict where each key is a B64 index and each value is the B64 char
#  B64IdxByChr is dict where each key is a B64 char and each value is the B64 index
# Map Base64 index to char
B64ChrByIdx = dict((index, char) for index, char in enumerate([chr(x) for x in range(65, 91)]))
B64ChrByIdx.update([(index + 26, char) for index, char in enumerate([chr(x) for x in range(97, 123)])])
B64ChrByIdx.update([(index + 52, char) for index, char in enumerate([chr(x) for x in range(48, 58)])])
B64ChrByIdx[62] = '-'
B64ChrByIdx[63] = '_'
# Map char to Base64 index
B64IdxByChr = {char: index for index, char in B64ChrByIdx.items()}

B64REX = rb'^[0-9A-Za-z_-]*\Z'   # [A-Za-z0-9\-\_]  bytes
# Usage: if Reb64.match(bext): or if not Reb64.match(bext): bext is bytes
Reb64 = re.compile(B64REX)  # compile is faster

def intToB64(i, l=1):
    """Converts int i to at least l Base64 chars as str.
    Returns:
        b64 (str): Base64 converstion of i of length minimum l. If more than
                    l chars are needed to represent i in Base64 then returned
                    str is lengthed appropriately. When less then l chars is
                    needed then returned str is prepadded with 'A' character.

    l is min number of b64 digits left padded with Base64 0 == "A" char
    The length of return bytes extended to accomodate full Base64 encoding of i
    """
    d = deque()  # deque of characters base64

    while l:
        d.appendleft(B64ChrByIdx[i % 64])
        i = i // 64
        if not i:
            break
    for j in range(l - len(d)):  # range(x)  x <= 0 means do not iterate
        d.appendleft("A")
    return ("".join(d))


def intToB64b(i, l=1):
    """Converts int i to at least l Base64 chars as bytes.

    Returns:
        b64 (bytes): Base64 converstion of i of length minimum l. If more than
                     l bytes are needed to represent i in Base64 then returned
                     bytes is extended appropriately. When less then l bytes
                     is needed then returned bytes is prepadded with b'A' bytes.

    Parameters:
        i (int): to be converted
        l (int): min number of b64 digits. When empty these are left padded with
                 Base64 0 == b'A' digits.
                 The length of return bytes is extended to accomodate full
                 Base64 encoding of i regardless of l.
    """
    return (intToB64(i=i, l=l).encode())


def b64ToInt(s):
    """Converts Base64 to int
    Returns:
        i (int): conversion s (str | bytes) to int

    Parameters:
        s (str | bytes): to be converted
    """
    if not s:
        raise ValueError("Empty string, conversion undefined.")
    if hasattr(s, 'decode'):
        s = s.decode("utf-8")
    i = 0
    for e, c in enumerate(reversed(s)):
        i |= B64IdxByChr[c] << (e * 6)  # same as i += B64IdxByChr[c] * (64 ** e)
    return i



def codeB64ToB2(s):
    """Convert Base64 chars in s to B2 bytes

    Returns:
        bs (bytes): conversion (decode) of s Base64 chars to Base2 bytes.
        Where the number of total bytes returned is equal to the minimun number of
        chars (octet) sufficient to hold the total converted concatenated chars from s,
        with one sextet per each Base64 char of s. Assumes no pad chars in s.


    Sextets are left aligned with pad bits in last (rightmost) byte to support
    mid padding of code portion with respect to rest of primitive.
    This is useful for decoding as bytes, code characters from the front of
    a Base64 encoded string of characters.

    Parameters:
        s (str | bytes): Base64 str or bytes to convert

    """
    i = b64ToInt(s)
    i <<= 2 * (len(s) % 4)  # add 2 bits right zero padding for each sextet
    n = sceil(len(s) * 3 / 4)  # compute min number of ocetets to hold all sextets
    return (i.to_bytes(n, 'big'))


def codeB2ToB64(b, l):
    """Convert l sextets from base2 b to base64 str

    Returns:
        code (str): conversion (encode) of l Base2 sextets from front of b
        to Base64 chars.

    One char for each of l sextets from front (left) of b.
    This is useful for encoding as code characters, sextets from the front of
    a Base2 bytes (byte string). Must provide l because of ambiguity between l=3
    and l=4. Both require 3 bytes in b. Trailing pad bits are removed so
    returned sextets as characters are right aligned .

    Parameters:
        b (bytes | str): target from which to nab sextets
        l (int): number of sextets to convert from front of b
    """
    if hasattr(b, 'encode'):
        b = b.encode("utf-8")  # convert to bytes
    n = sceil(l * 3 / 4)  # number of bytes needed for l sextets
    if n > len(b):
        raise ValueError("Not enough bytes in {} to nab {} sextets.".format(b, l))
    i = int.from_bytes(b[:n], 'big')  # convert only first n bytes to int
    # check if prepad bits are zero
    tbs = 2 * (l % 4)  # trailing bit size in bits
    i >>= tbs  # right shift out trailing bits to make right aligned
    return (intToB64(i, l))  # return as B64


def nabSextets(b, l):
    """Nab l sextets from front of b
    Returns:
        sextets (bytes): first l sextets from front (left) of b as bytes
        (byte string). Length of bytes returned is minimum sufficient to hold
        all l sextets. Last byte returned is right bit padded with zeros which
        is compatible with mid padded codes on front of primitives

    Parameters:
        b (bytes | str): target from which to nab sextets
        l (int): number of sextets to nab from front of b

    """
    if hasattr(b, 'encode'):
        b = b.encode()  # convert to bytes
    n = sceil(l * 3 / 4)  # number of bytes needed for l sextets
    if n > len(b):
        raise ValueError("Not enough bytes in {} to nab {} sextets.".format(b, l))
    i = int.from_bytes(b[:n], 'big')
    p = 2 * (l % 4)
    i >>= p  # strip of last bits
    i <<= p  # pad with empty bits
    return (i.to_bytes(n, 'big'))


# Miscellaneous


# Regular expression to detect valid attribute names for Boxes
ATREX = r'^[a-zA-Z_][a-zA-Z0-9_]*$'
# Usage: if Reat.match(name): or if not Reat.match(name):
Reat = re.compile(ATREX)  # compile is faster


